@@ -34,7 +34,6 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <stdarg.h>
-#include <assert.h>
 
 
 // The global variables below comprise all mutable global data in GLFW
@@ -58,7 +57,10 @@ static _GLFWinitconfig _glfwInitHints = {
     .ns = {
         .menubar = true,  // macOS menu bar
         .chdir = true   // macOS bundle chdir
-    }
+    },
+    .wl = {
+        .ime = true,  // Wayland IME support
+    },
 };
 
 // Terminate the library
@@ -208,7 +210,7 @@ _glfwDebug(const char *format, ...) {
     {
         va_list vl;
 
-        fprintf(stderr, "[%.4f] ", monotonic_t_to_s_double(glfwGetTime()));
+        fprintf(stderr, "[%.3f] ", monotonic_t_to_s_double(monotonic()));
         va_start(vl, format);
         vfprintf(stderr, format, vl);
         va_end(vl);
@@ -221,8 +223,9 @@ _glfwDebug(const char *format, ...) {
 //////                        GLFW public API                       //////
 //////////////////////////////////////////////////////////////////////////
 
-GLFWAPI int glfwInit(monotonic_t start_time)
+GLFWAPI int glfwInit(monotonic_t start_time, bool *supports_window_occlusion)
 {
+    *supports_window_occlusion = false;
     if (_glfw.initialized)
         return true;
     monotonic_start_time = start_time;
@@ -231,7 +234,7 @@ GLFWAPI int glfwInit(monotonic_t start_time)
     _glfw.hints.init = _glfwInitHints;
     _glfw.ignoreOSKeyboardProcessing = false;
 
-    if (!_glfwPlatformInit())
+    if (!_glfwPlatformInit(supports_window_occlusion))
     {
         terminate();
         return false;
@@ -297,6 +300,9 @@ GLFWAPI void glfwInitHint(int hint, int value)
         case GLFW_COCOA_MENUBAR:
             _glfwInitHints.ns.menubar = value;
             return;
+        case GLFW_WAYLAND_IME:
+            _glfwInitHints.wl.ime = value;
+            return;
     }
 
     _glfwInputError(GLFW_INVALID_ENUM,
@@ -382,6 +388,14 @@ GLFWAPI GLFWapplicationclosefun glfwSetApplicationCloseCallback(GLFWapplicationc
     return cbfun;
 }
 
+GLFWAPI GLFWsystemcolorthemechangefun glfwSetSystemColorThemeChangeCallback(GLFWsystemcolorthemechangefun cbfun)
+{
+    _GLFW_REQUIRE_INIT_OR_RETURN(NULL);
+    _GLFW_SWAP_POINTERS(_glfw.callbacks.system_color_theme_change, cbfun);
+    return cbfun;
+}
+
+
 GLFWAPI GLFWdrawtextfun glfwSetDrawTextFunction(GLFWdrawtextfun cbfun)
 {
     _GLFW_REQUIRE_INIT_OR_RETURN(NULL);
@@ -402,3 +416,10 @@ GLFWAPI GLFWhascurrentselectionfun glfwSetHasCurrentSelectionCallback(GLFWhascur
     _GLFW_SWAP_POINTERS(_glfw.callbacks.has_current_selection, cbfun);
     return cbfun;
 }
+
+GLFWAPI GLFWimecursorpositionfun glfwSetIMECursorPositionCallback(GLFWimecursorpositionfun cbfun)
+{
+    _GLFW_REQUIRE_INIT_OR_RETURN(NULL);
+    _GLFW_SWAP_POINTERS(_glfw.callbacks.get_ime_cursor_position, cbfun);
+    return cbfun;
+}
