@@ -44,6 +44,8 @@
 #include <sys/socket.h>
 #include <wayland-client.h>
 #include <stdio.h>
+// errno.h needed for BSD code paths
+#include <errno.h>
 // Needed for the BTN_* defines
 #ifdef __has_include
 #if __has_include(<linux/input.h>)
@@ -92,7 +94,6 @@ pointerHandleEnter(
         window->wl.hovered = true;
         window->wl.cursorPosX = x;
         window->wl.cursorPosY = y;
-
         _glfwPlatformSetCursor(window, window->wl.currentCursor);
         _glfwInputCursorEnter(window, true);
     }
@@ -222,9 +223,9 @@ pointer_handle_frame(void *data UNUSED, struct wl_pointer *pointer UNUSED) {
     memset(&info, 0, sizeof(info));
 
     if (x != 0.0f || y != 0.0f) {
-        float scale = _glfwWaylandWindowScale(window);
+        float scale = (float)_glfwWaylandWindowScale(window);
         y *= scale; x *= scale;
-        _glfwInputScroll(window, x, y, highres, _glfw.wl.xkb.states.modifiers);
+        _glfwInputScroll(window, -x, y, highres, _glfw.wl.xkb.states.modifiers);
     }
 }
 
@@ -521,7 +522,9 @@ static void registryHandleGlobal(void* data UNUSED,
     else if (is(xdg_wm_base))
     {
         _glfw.wl.xdg_wm_base_version = 1;
-#ifdef XDG_TOPLEVEL_WM_CAPABILITIES_SINCE_VERSION
+#ifdef XDG_TOPLEVEL_STATE_SUSPENDED_SINCE_VERSION
+        _glfw.wl.xdg_wm_base_version = MIN(XDG_TOPLEVEL_STATE_SUSPENDED_SINCE_VERSION, (int)version);
+#elif defined(XDG_TOPLEVEL_WM_CAPABILITIES_SINCE_VERSION)
         _glfw.wl.xdg_wm_base_version = MIN(XDG_TOPLEVEL_WM_CAPABILITIES_SINCE_VERSION, (int)version);
 #endif
         _glfw.wl.wmBase = wl_registry_bind(registry, name, &xdg_wm_base_interface, _glfw.wl.xdg_wm_base_version);
@@ -595,6 +598,9 @@ static void registryHandleGlobal(void* data UNUSED,
             _glfw.wl.zwlr_layer_shell_v1 = wl_registry_bind(registry, name, &zwlr_layer_shell_v1_interface, version);
         }
     }
+    else if (is(zwp_idle_inhibit_manager_v1)) {
+        _glfw.wl.idle_inhibit_manager = wl_registry_bind(registry, name, &zwp_idle_inhibit_manager_v1_interface, 1);
+    }
 #undef is
 }
 
@@ -629,12 +635,13 @@ static const struct wl_registry_listener registryListener = {
 };
 
 
-GLFWAPI GLFWColorScheme glfwGetCurrentSystemColorTheme(void) {
-    return glfw_current_system_color_theme();
+GLFWAPI GLFWColorScheme glfwGetCurrentSystemColorTheme(bool query_if_unintialized) {
+    return glfw_current_system_color_theme(query_if_unintialized);
 }
 
 static pid_t
 get_socket_peer_pid(int fd) {
+    (void)fd;
 #ifdef __linux__
     struct ucred ucred;
     socklen_t len = sizeof(struct ucred);
@@ -660,13 +667,42 @@ GLFWAPI pid_t glfwWaylandCompositorPID(void) {
     return get_socket_peer_pid(fd);
 }
 
+const char*
+_glfwWaylandCompositorName(void) {
+    static bool probed = false;
+    if (!probed) {
+        probed = true;
+        static const size_t sz = 1024;
+        _glfw.wl.compositor_name = malloc(sz);
+        if (!_glfw.wl.compositor_name) return "";
+        char *ans = _glfw.wl.compositor_name; ans[0] = 0;
+        pid_t cpid = glfwWaylandCompositorPID();
+        if (cpid < 0) return ans;
+        snprintf(ans, sz, "/proc/%d/cmdline", cpid);
+        int fd = open(ans, O_RDONLY | O_CLOEXEC);
+        if (fd < 0) {
+            ans[0] = 0;
+        } else {
+            ssize_t n;
+            while (true) {
+                n = read(fd, ans, sz-1);
+                if (n < 0 && errno == EINTR) continue;
+                close(fd); break;
+            }
+            ans[n < 0 ? 0 : n] = 0;
+        }
+    }
+    return _glfw.wl.compositor_name ? _glfw.wl.compositor_name : "";
+}
+
+
 //////////////////////////////////////////////////////////////////////////
 //////                       GLFW platform API                      //////
 //////////////////////////////////////////////////////////////////////////
 
 static const char*
 get_compositor_missing_capabilities(void) {
-#define C(title, x) if (!_glfw.wl.x) p += snprintf(buf, sizeof(buf) - (p - buf), "%s", #title);
+#define C(title, x) if (!_glfw.wl.x) p += snprintf(buf, sizeof(buf) - (p - buf), "%s ", #title);
     static char buf[256];
     char *p = buf;
     *p = 0;
@@ -674,13 +710,17 @@ get_compositor_missing_capabilities(void) {
     C(blur, org_kde_kwin_blur_manager); C(server_side_decorations, decorationManager);
     C(cursor_shape, wp_cursor_shape_manager_v1); C(layer_shell, zwlr_layer_shell_v1);
     C(single_pixel_buffer, wp_single_pixel_buffer_manager_v1); C(preferred_scale, has_preferred_buffer_scale);
+    C(idle_inhibit, idle_inhibit_manager);
+    if (_glfw.wl.xdg_wm_base_version < 6) p += snprintf(buf, sizeof(buf) - (p - buf), "%s ", "window-state-suspended");
+    if (_glfw.wl.xdg_wm_base_version < 5) p += snprintf(buf, sizeof(buf) - (p - buf), "%s ", "window-capabilities");
 #undef C
+    while (p > buf && (p - 1)[0] == ' ') { p--; *p = 0; }
     return buf;
 }
 
 GLFWAPI const char* glfwWaylandMissingCapabilities(void) { return get_compositor_missing_capabilities(); }
 
-int _glfwPlatformInit(void)
+int _glfwPlatformInit(bool *supports_window_occlusion)
 {
     int i;
     _GLFWmonitor* monitor;
@@ -771,6 +811,7 @@ int _glfwPlatformInit(void)
         const char *mc = get_compositor_missing_capabilities();
         if (mc && mc[0]) debug("Compositor missing capabilities: %s\n", mc);
     }
+    *supports_window_occlusion = _glfw.wl.xdg_wm_base_version > 5;
 
     return true;
 }
@@ -856,6 +897,8 @@ void _glfwPlatformTerminate(void)
         org_kde_kwin_blur_manager_destroy(_glfw.wl.org_kde_kwin_blur_manager);
     if (_glfw.wl.zwlr_layer_shell_v1)
         zwlr_layer_shell_v1_destroy(_glfw.wl.zwlr_layer_shell_v1);
+    if (_glfw.wl.idle_inhibit_manager)
+        zwp_idle_inhibit_manager_v1_destroy(_glfw.wl.idle_inhibit_manager);
 
     if (_glfw.wl.registry)
         wl_registry_destroy(_glfw.wl.registry);
@@ -863,8 +906,13 @@ void _glfwPlatformTerminate(void)
     {
         wl_display_flush(_glfw.wl.display);
         wl_display_disconnect(_glfw.wl.display);
+        _glfw.wl.display = NULL;
     }
     finalizePollData(&_glfw.wl.eventLoopData);
+    if (_glfw.wl.compositor_name) {
+        free(_glfw.wl.compositor_name);
+        _glfw.wl.compositor_name = NULL;
+    }
 }
 
 #define GLFW_LOOP_BACKEND wl
