@@ -4,8 +4,10 @@
 import os
 import sys
 from collections import defaultdict
+from collections.abc import Generator, Sequence
 from contextlib import contextmanager, suppress
-from typing import TYPE_CHECKING, DefaultDict, Dict, Generator, List, Optional, Sequence, Tuple
+from itertools import count
+from typing import TYPE_CHECKING, DefaultDict, Iterable, Optional, TypedDict
 
 import kitty.fast_data_types as fast_data_types
 
@@ -13,10 +15,6 @@ from .constants import handled_signals, is_freebsd, is_macos, kitten_exe, kitty_
 from .types import run_once
 from .utils import cmdline_for_hold, log_error, which
 
-try:
-    from typing import TypedDict
-except ImportError:
-    TypedDict = dict
 if TYPE_CHECKING:
     from .window import CwdRequest
 
@@ -28,19 +26,21 @@ if is_macos:
     from kitty.fast_data_types import process_group_map as _process_group_map
 
     def cwd_of_process(pid: int) -> str:
+        # The underlying code on macos returns a path with symlinks resolved
+        # anyway but we use realpath for extra safety.
         return os.path.realpath(_cwd(pid))
 
-    def process_group_map() -> DefaultDict[int, List[int]]:
-        ans: DefaultDict[int, List[int]] = defaultdict(list)
+    def process_group_map() -> DefaultDict[int, list[int]]:
+        ans: DefaultDict[int, list[int]] = defaultdict(list)
         for pid, pgid in _process_group_map():
             ans[pgid].append(pid)
         return ans
 
-    def cmdline_of_pid(pid: int) -> List[str]:
+    def cmdline_of_pid(pid: int) -> list[str]:
         return cmdline_(pid)
 else:
 
-    def cmdline_of_pid(pid: int) -> List[str]:
+    def cmdline_of_pid(pid: int) -> list[str]:
         with open(f'/proc/{pid}/cmdline', 'rb') as f:
             return list(filter(None, f.read().decode('utf-8').split('\0')))
 
@@ -54,6 +54,8 @@ else:
             return os.path.realpath(ans)
     else:
         def cwd_of_process(pid: int) -> str:
+            # We use realpath instead of readlink to match macOS behavior where
+            # the underlying OS API returns real paths.
             ans = f'/proc/{pid}/cwd'
             return os.path.realpath(ans)
 
@@ -61,8 +63,8 @@ else:
         with open(f'/proc/{pid}/environ', 'rb') as f:
             return f.read().decode('utf-8')
 
-    def process_group_map() -> DefaultDict[int, List[int]]:
-        ans: DefaultDict[int, List[int]] = defaultdict(list)
+    def process_group_map() -> DefaultDict[int, list[int]]:
+        ans: DefaultDict[int, list[int]] = defaultdict(list)
         for x in os.listdir('/proc'):
             try:
                 pid = int(x)
@@ -82,12 +84,12 @@ else:
 
 
 @run_once
-def checked_terminfo_dir() -> Optional[str]:
+def checked_terminfo_dir() -> str | None:
     return terminfo_dir if os.path.isdir(terminfo_dir) else None
 
 
-def processes_in_group(grp: int) -> List[int]:
-    gmap: Optional[DefaultDict[int, List[int]]] = getattr(process_group_map, 'cached_map', None)
+def processes_in_group(grp: int) -> list[int]:
+    gmap: DefaultDict[int, list[int]] | None = getattr(process_group_map, 'cached_map', None)
     if gmap is None:
         try:
             gmap = process_group_map()
@@ -109,11 +111,18 @@ def cached_process_data() -> Generator[None, None, None]:
         delattr(process_group_map, 'cached_map')
 
 
-def parse_environ_block(data: str) -> Dict[str, str]:
+def session_id(pids: Iterable[int]) -> int:
+    for pid in pids:
+        with suppress(OSError):
+            if (sid := os.getsid(pid)) > -1:
+                return sid
+    return -1
+
+def parse_environ_block(data: str) -> dict[str, str]:
     """Parse a C environ block of environment variables into a dictionary."""
     # The block is usually raw data from the target process.  It might contain
     # trailing garbage and lines that do not look like assignments.
-    ret: Dict[str, str] = {}
+    ret: dict[str, str] = {}
     pos = 0
 
     while True:
@@ -132,27 +141,28 @@ def parse_environ_block(data: str) -> Dict[str, str]:
     return ret
 
 
-def environ_of_process(pid: int) -> Dict[str, str]:
+def environ_of_process(pid: int) -> dict[str, str]:
     return parse_environ_block(_environ_of_process(pid))
 
 
-def process_env() -> Dict[str, str]:
+def process_env() -> dict[str, str]:
     ans = dict(os.environ)
     ssl_env_var = getattr(sys, 'kitty_ssl_env_var', None)
     if ssl_env_var is not None:
         ans.pop(ssl_env_var, None)
     ans.pop('XDG_ACTIVATION_TOKEN', None)
+    ans.pop('VTE_VERSION', None)  # Used by the stupid VTE shell integration script that is installed system wide, sigh
     return ans
 
 
-def default_env() -> Dict[str, str]:
-    ans: Optional[Dict[str, str]] = getattr(default_env, 'env', None)
+def default_env() -> dict[str, str]:
+    ans: dict[str, str] | None = getattr(default_env, 'env', None)
     if ans is None:
         return process_env()
     return ans
 
 
-def set_default_env(val: Optional[Dict[str, str]] = None) -> None:
+def set_default_env(val: dict[str, str] | None = None) -> None:
     env = process_env().copy()
     has_lctype = False
     if val:
@@ -166,7 +176,7 @@ def set_LANG_in_default_env(val: str) -> None:
     default_env().setdefault('LANG', val)
 
 
-def openpty() -> Tuple[int, int]:
+def openpty() -> tuple[int, int]:
     master, slave = os.openpty()  # Note that master and slave are in blocking mode
     os.set_inheritable(slave, True)
     os.set_inheritable(master, False)
@@ -185,31 +195,39 @@ def base64_terminfo_data() -> str:
 
 
 class ProcessDesc(TypedDict):
-    cwd: Optional[str]
+    cwd: str | None
     pid: int
-    cmdline: Optional[Sequence[str]]
+    cmdline: Sequence[str] | None
+
+
+child_counter = count()
 
 
 class Child:
 
-    child_fd: Optional[int] = None
-    pid: Optional[int] = None
+    child_fd: int | None = None
+    pid: int | None = None
     forked = False
 
     def __init__(
         self,
         argv: Sequence[str],
         cwd: str,
-        stdin: Optional[bytes] = None,
-        env: Optional[Dict[str, str]] = None,
+        stdin: bytes | None = None,
+        env: dict[str, str] | None = None,
         cwd_from: Optional['CwdRequest'] = None,
         is_clone_launch: str = '',
         add_listen_on_env_var: bool = True,
         hold: bool = False,
+        pass_fds: tuple[int, ...] = (),
+        remote_control_fd: int = -1,
     ):
         self.is_clone_launch = is_clone_launch
+        self.id = next(child_counter)
         self.add_listen_on_env_var = add_listen_on_env_var
         self.argv = list(argv)
+        self.pass_fds = pass_fds
+        self.remote_control_fd = remote_control_fd
         if cwd_from:
             try:
                 cwd = cwd_from.modify_argv_for_launch_with_cwd(self.argv, env) or cwd
@@ -220,12 +238,12 @@ class Child:
         self.cwd = os.path.abspath(cwd)
         self.stdin = stdin
         self.env = env or {}
-        self.final_env:Dict[str, str] = {}
+        self.final_env:dict[str, str] = {}
         self.is_default_shell = bool(self.argv and self.argv[0] == shell_path)
         self.should_run_via_run_shell_kitten = is_macos and self.is_default_shell
         self.hold = hold
 
-    def get_final_env(self) -> Dict[str, str]:
+    def get_final_env(self) -> dict[str, str]:
         from kitty.options.utils import DELETE_ENV_VAR
         env = default_env().copy()
         opts = fast_data_types.get_options()
@@ -238,10 +256,13 @@ class Child:
         env['COLORTERM'] = 'truecolor'
         env['KITTY_PID'] = getpid()
         env['KITTY_PUBLIC_KEY'] = boss.encryption_public_key
-        if self.add_listen_on_env_var and boss.listening_on:
+        if self.remote_control_fd > -1:
+            env['KITTY_LISTEN_ON'] = f'fd:{self.remote_control_fd}'
+        elif self.add_listen_on_env_var and boss.listening_on:
             env['KITTY_LISTEN_ON'] = boss.listening_on
         else:
             env.pop('KITTY_LISTEN_ON', None)
+        env.pop('KITTY_STDIO_FORWARDED', None)
         if self.cwd:
             # needed in case cwd is a symlink, in which case shells
             # can use it to display the current directory name rather
@@ -254,8 +275,6 @@ class Child:
         elif opts.terminfo_type == 'direct':
             env['TERMINFO'] = base64_terminfo_data()
         env['KITTY_INSTALLATION_DIR'] = kitty_base_dir
-        if opts.forward_stdio:
-            env['KITTY_STDIO_FORWARDED'] = '3'
         self.unmodified_argv = list(self.argv)
         if not self.should_run_via_run_shell_kitten and 'disabled' not in opts.shell_integration:
             from .shell_integration import modify_shell_environ
@@ -268,7 +287,7 @@ class Child:
             env.pop('KITTY_IS_CLONE_LAUNCH', None)
         return env
 
-    def fork(self) -> Optional[int]:
+    def fork(self) -> int | None:
         if self.forked:
             return None
         opts = fast_data_types.get_options()
@@ -287,6 +306,9 @@ class Child:
         self.final_env = self.get_final_env()
         argv = list(self.argv)
         cwd = self.cwd
+        pass_fds = self.pass_fds
+        if self.remote_control_fd > -1:
+            pass_fds += self.remote_control_fd,
         if self.should_run_via_run_shell_kitten:
             # bash will only source ~/.bash_profile if it detects it is a login
             # shell (see the invocation section of the bash man page), which it
@@ -307,12 +329,14 @@ class Child:
             if ksi == 'invalid':
                 ksi = 'enabled'
             argv = [kitten_exe(), 'run-shell', '--shell', shlex.join(argv), '--shell-integration', ksi]
-            if is_macos:
+            if is_macos and not pass_fds and not opts.forward_stdio:
                 # In addition for getlogin() to work we need to run the shell
                 # via the /usr/bin/login wrapper, sigh.
                 # And login on macOS looks for .hushlogin in CWD instead of
                 # HOME, bloody idiotic so we cant cwd when running it.
                 # https://github.com/kovidgoyal/kitty/issues/6511
+                # login closes inherited file descriptors so dont use it when
+                # forward_stdio or pass_fds are used.
                 import pwd
                 user = pwd.getpwuid(os.geteuid()).pw_name
                 if cwd:
@@ -327,7 +351,7 @@ class Child:
         env = tuple(f'{k}={v}' for k, v in self.final_env.items())
         pid = fast_data_types.spawn(
             final_exe, cwd, tuple(argv), env, master, slave, stdin_read_fd, stdin_write_fd,
-            ready_read_fd, ready_write_fd, tuple(handled_signals), kitten_exe(), opts.forward_stdio)
+            ready_read_fd, ready_write_fd, tuple(handled_signals), kitten_exe(), opts.forward_stdio, pass_fds)
         os.close(slave)
         self.pid = pid
         self.child_fd = master
@@ -338,6 +362,14 @@ class Child:
         self.terminal_ready_fd = ready_write_fd
         if self.child_fd is not None:
             os.set_blocking(self.child_fd, False)
+        if not is_macos:
+            ppid = getpid()
+            try:
+                fast_data_types.systemd_move_pid_into_new_scope(pid, f'kitty-{ppid}-{self.id}.scope', f'kitty child process: {pid} launched by: {ppid}')
+            except NotImplementedError:
+                pass
+            except OSError as err:
+                log_error("Could not move child process into a systemd scope: " + str(err))
         return pid
 
     def __del__(self) -> None:
@@ -350,7 +382,7 @@ class Child:
         os.close(self.terminal_ready_fd)
         self.terminal_ready_fd = -1
 
-    def cmdline_of_pid(self, pid: int) -> List[str]:
+    def cmdline_of_pid(self, pid: int) -> list[str]:
         try:
             ans = cmdline_of_pid(pid)
         except Exception:
@@ -359,28 +391,51 @@ class Child:
             ans = list(self.argv)
         return ans
 
+    def process_desc(self, pid: int) -> ProcessDesc:
+        ans: ProcessDesc = {'pid': pid, 'cmdline': None, 'cwd': None}
+        with suppress(Exception):
+            ans['cmdline'] = self.cmdline_of_pid(pid)
+        with suppress(Exception):
+            ans['cwd'] = cwd_of_process(pid) or None
+        return ans
+
     @property
-    def foreground_processes(self) -> List[ProcessDesc]:
+    def foreground_processes(self) -> list[ProcessDesc]:
         if self.child_fd is None:
             return []
         try:
             pgrp = os.tcgetpgrp(self.child_fd)
             foreground_processes = processes_in_group(pgrp) if pgrp >= 0 else []
-
-            def process_desc(pid: int) -> ProcessDesc:
-                ans: ProcessDesc = {'pid': pid, 'cmdline': None, 'cwd': None}
-                with suppress(Exception):
-                    ans['cmdline'] = self.cmdline_of_pid(pid)
-                with suppress(Exception):
-                    ans['cwd'] = cwd_of_process(pid) or None
-                return ans
-
-            return [process_desc(x) for x in foreground_processes]
+            return [self.process_desc(x) for x in foreground_processes]
         except Exception:
             return []
 
     @property
-    def cmdline(self) -> List[str]:
+    def background_processes(self) -> list[ProcessDesc]:
+        if self.child_fd is None:
+            return []
+        try:
+            foreground_process_group_id = os.tcgetpgrp(self.child_fd)
+            if foreground_process_group_id < 0:
+                return []
+            gmap = process_group_map()
+
+            sid = session_id(gmap.get(foreground_process_group_id, ()))
+            if sid < 0:
+                return []
+            ans = []
+            for grp_id, pids in gmap.items():
+                if grp_id == foreground_process_group_id:
+                    continue
+                if session_id(pids) == sid:
+                    for pid in pids:
+                        ans.append(self.process_desc(pid))
+            return ans
+        except Exception:
+            return []
+
+    @property
+    def cmdline(self) -> list[str]:
         try:
             assert self.pid is not None
             return self.cmdline_of_pid(self.pid) or list(self.argv)
@@ -388,7 +443,7 @@ class Child:
             return list(self.argv)
 
     @property
-    def foreground_cmdline(self) -> List[str]:
+    def foreground_cmdline(self) -> list[str]:
         try:
             assert self.pid_for_cwd is not None
             return self.cmdline_of_pid(self.pid_for_cwd) or self.cmdline
@@ -396,7 +451,7 @@ class Child:
             return self.cmdline
 
     @property
-    def environ(self) -> Dict[str, str]:
+    def environ(self) -> dict[str, str]:
         try:
             assert self.pid is not None
             return environ_of_process(self.pid) or self.final_env.copy()
@@ -404,13 +459,13 @@ class Child:
             return self.final_env.copy()
 
     @property
-    def current_cwd(self) -> Optional[str]:
+    def current_cwd(self) -> str | None:
         with suppress(Exception):
             assert self.pid is not None
             return cwd_of_process(self.pid)
         return None
 
-    def get_pid_for_cwd(self, oldest: bool = False) -> Optional[int]:
+    def get_pid_for_cwd(self, oldest: bool = False) -> int | None:
         with suppress(Exception):
             assert self.child_fd is not None
             pgrp = os.tcgetpgrp(self.child_fd)
@@ -430,17 +485,17 @@ class Child:
         return self.pid
 
     @property
-    def pid_for_cwd(self) -> Optional[int]:
+    def pid_for_cwd(self) -> int | None:
         return self.get_pid_for_cwd()
 
-    def get_foreground_cwd(self, oldest: bool = False) -> Optional[str]:
+    def get_foreground_cwd(self, oldest: bool = False) -> str | None:
         with suppress(Exception):
             pid = self.get_pid_for_cwd(oldest)
             if pid is not None:
                 return cwd_of_process(pid) or None
         return None
 
-    def get_foreground_exe(self, oldest: bool = False) -> Optional[str]:
+    def get_foreground_exe(self, oldest: bool = False) -> str | None:
         with suppress(Exception):
             pid = self.get_pid_for_cwd(oldest)
             if pid is not None:
@@ -450,11 +505,11 @@ class Child:
         return None
 
     @property
-    def foreground_cwd(self) -> Optional[str]:
+    def foreground_cwd(self) -> str | None:
         return self.get_foreground_cwd()
 
     @property
-    def foreground_environ(self) -> Dict[str, str]:
+    def foreground_environ(self) -> dict[str, str]:
         pid = self.pid_for_cwd
         if pid is not None:
             with suppress(Exception):
@@ -475,7 +530,7 @@ class Child:
             return False
         cc = t[-1]
         if key_num == cc[termios.VINTR]:
-            s = signal.SIGINT
+            s: signal.Signals = signal.SIGINT
         elif key_num == cc[termios.VSUSP]:
             s = signal.SIGTSTP
         elif key_num == cc[termios.VQUIT]:
