@@ -13,6 +13,7 @@
 #endif
 
 #include "data-types.h"
+#include "line.h"
 #include "charsets.h"
 #include "base64.h"
 #include <sys/socket.h>
@@ -41,14 +42,14 @@
 #include <xlocale.h>
 
 static PyObject*
-user_cache_dir(void) {
+user_cache_dir(PyObject *self UNUSED, PyObject *args UNUSED) {
     static char buf[1024];
     if (!confstr(_CS_DARWIN_USER_CACHE_DIR, buf, sizeof(buf) - 1)) return PyErr_SetFromErrno(PyExc_OSError);
     return PyUnicode_FromString(buf);
 }
 
 static PyObject*
-process_group_map(void) {
+process_group_map(PyObject *self UNUSED, PyObject *args UNUSED) {
     int num_of_processes = proc_listallpids(NULL, 0);
     size_t bufsize = sizeof(pid_t) * (num_of_processes + 1024);
     RAII_ALLOC(pid_t, buf, malloc(bufsize));
@@ -77,10 +78,13 @@ redirect_std_streams(PyObject UNUSED *self, PyObject *args) {
 }
 
 static PyObject*
-pybase64_encode(PyObject UNUSED *self, PyObject *args) {
+pybase64_encode(PyObject UNUSED *self, PyObject *const *args, Py_ssize_t nargs) {
     int add_padding = 0;
+    if (nargs < 1 || nargs > 2) { PyErr_SetString(PyExc_TypeError, "must supply one or two arguments"); return NULL; }
     RAII_PY_BUFFER(view);
-    if (!PyArg_ParseTuple(args, "s*|p", &view, &add_padding)) return NULL;
+    if (PyUnicode_Check(args[0])) view.buf = (void*)PyUnicode_AsUTF8AndSize(args[0], &view.len);
+    else if (PyObject_GetBuffer(args[0], &view, PyBUF_SIMPLE) != 0) return NULL;
+    if (nargs == 2) add_padding = PyObject_IsTrue(args[1]);
     size_t sz = required_buffer_size_for_base64_encode(view.len);
     PyObject *ans = PyBytes_FromStringAndSize(NULL, sz);
     if (!ans) return NULL;
@@ -90,9 +94,21 @@ pybase64_encode(PyObject UNUSED *self, PyObject *args) {
 }
 
 static PyObject*
-pybase64_decode(PyObject UNUSED *self, PyObject *args) {
+base64_encode_into(PyObject UNUSED *self, PyObject *args) {
+    int add_padding = 0;
+    RAII_PY_BUFFER(view); RAII_PY_BUFFER(output);
+    if (!PyArg_ParseTuple(args, "s*w*|i", &view, &output, &add_padding)) return NULL;
+    size_t sz = required_buffer_size_for_base64_encode(view.len);
+    if (output.len < (ssize_t)sz) { PyErr_SetString(PyExc_TypeError, "output buffer too small"); return NULL; }
+    base64_encode8(view.buf, view.len, output.buf, &sz, add_padding);
+    return PyLong_FromSize_t(sz);
+}
+
+static PyObject*
+pybase64_decode(PyObject UNUSED *self, PyObject *input_data) {
     RAII_PY_BUFFER(view);
-    if (!PyArg_ParseTuple(args, "s*", &view)) return NULL;
+    if (PyUnicode_Check(input_data)) view.buf = (void*)PyUnicode_AsUTF8AndSize(input_data, &view.len);
+    else if (PyObject_GetBuffer(input_data, &view, PyBUF_SIMPLE) != 0) return NULL;
     size_t sz = required_buffer_size_for_base64_decode(view.len);
     PyObject *ans = PyBytes_FromStringAndSize(NULL, sz);
     if (!ans) return NULL;
@@ -105,6 +121,186 @@ pybase64_decode(PyObject UNUSED *self, PyObject *args) {
     return ans;
 }
 
+static PyObject*
+base64_decode_into(PyObject UNUSED *self, PyObject *args) {
+    RAII_PY_BUFFER(view); RAII_PY_BUFFER(output);
+    if (!PyArg_ParseTuple(args, "s*w*", &view, &output)) return NULL;
+    size_t sz = required_buffer_size_for_base64_decode(view.len);
+    if (output.len < (ssize_t)sz) { PyErr_SetString(PyExc_TypeError, "output buffer too small"); return NULL; }
+    if (!base64_decode8(view.buf, view.len, output.buf, &sz)) {
+        PyErr_SetString(PyExc_ValueError, "Invalid base64 input data");
+        return NULL;
+    }
+    return PyLong_FromSize_t(sz);
+}
+
+
+typedef struct StreamingBase64Decoder {
+    PyObject_HEAD
+    struct base64_state state;
+    bool add_trailing_bytes, needs_more_data;
+} StreamingBase64Decoder;
+
+static void
+StreamingBase64Decoder_reset_(StreamingBase64Decoder *self) {
+    base64_stream_decode_init(&self->state, 0);
+    self->needs_more_data = false;
+}
+
+static int
+StreamingBase64Decoder_init(PyObject *s, PyObject *args, PyObject *kwds UNUSED) {
+    if (PyTuple_GET_SIZE(args)) { PyErr_SetString(PyExc_TypeError, "constructor takes no arguments"); return -1; }
+    StreamingBase64Decoder *self = (StreamingBase64Decoder*)s;
+	base64_stream_decode_init(&self->state, 0);
+    return 0;
+}
+
+static PyObject*
+StreamingBase64Decoder_decode(StreamingBase64Decoder *self, PyObject *a) {
+    RAII_PY_BUFFER(data);
+    if (PyObject_GetBuffer(a, &data, PyBUF_SIMPLE) != 0) return NULL;
+    if (!data.buf || !data.len) return PyBytes_FromStringAndSize(NULL, 0);
+    size_t sz = required_buffer_size_for_base64_decode(data.len);
+    RAII_PyObject(ans, PyBytes_FromStringAndSize(NULL, sz));
+    if (!ans) return NULL;
+    int ret;
+    Py_BEGIN_ALLOW_THREADS
+    ret = base64_stream_decode(&self->state, data.buf, data.len, PyBytes_AS_STRING(ans), &sz);
+    Py_END_ALLOW_THREADS;
+    if (!ret) {
+        StreamingBase64Decoder_reset_(self);
+        PyErr_SetString(PyExc_ValueError, "Invalid base64 input data");
+        return NULL;
+    }
+    if (self->state.eof) StreamingBase64Decoder_reset_(self);
+    else self->needs_more_data = self->state.carry != 0 || self->state.bytes != 0;
+    if (_PyBytes_Resize(&ans, sz) != 0) return NULL;
+    return Py_NewRef(ans);
+}
+
+static PyObject*
+StreamingBase64Decoder_decode_into(StreamingBase64Decoder *self, PyObject *const *args, Py_ssize_t nargs) {
+    if (nargs != 2) { PyErr_SetString(PyExc_TypeError, "constructor takes exactly two arguments"); return NULL; }
+    RAII_PY_BUFFER(data);
+    if (PyObject_GetBuffer(args[0], &data, PyBUF_WRITE) != 0) return NULL;
+    if (!data.buf || !data.len) return PyLong_FromLong(0);
+    RAII_PY_BUFFER(src);
+    if (PyObject_GetBuffer(args[1], &src, PyBUF_SIMPLE) != 0) return NULL;
+    if (!src.buf || !src.len) return PyLong_FromLong(0);
+    size_t sz = required_buffer_size_for_base64_decode(src.len);
+    if ((Py_ssize_t)sz > data.len) { PyErr_SetString(PyExc_BufferError, "output buffer too small"); return NULL; }
+    int ret;
+    Py_BEGIN_ALLOW_THREADS
+    ret = base64_stream_decode(&self->state, src.buf, src.len, data.buf, &sz);
+    Py_END_ALLOW_THREADS
+    if (!ret) {
+        StreamingBase64Decoder_reset_(self);
+        PyErr_SetString(PyExc_ValueError, "Invalid base64 input data");
+        return NULL;
+    }
+    if (self->state.eof) StreamingBase64Decoder_reset_(self); else self->needs_more_data = true;
+    return PyLong_FromSize_t(sz);
+}
+
+static PyObject*
+StreamingBase64Decoder_reset(StreamingBase64Decoder *self, PyObject *args UNUSED) {
+    StreamingBase64Decoder_reset_(self);
+    Py_RETURN_NONE;
+}
+
+static PyObject*
+StreamingBase64Decoder_needs_more_data(StreamingBase64Decoder *self, PyObject *args UNUSED) {
+    return Py_NewRef(self->needs_more_data ? Py_True : Py_False);
+}
+
+static PyTypeObject StreamingBase64Decoder_Type = {
+    PyVarObject_HEAD_INIT(NULL, 0)
+    .tp_name = "kitty.fast_data_types.StreamingBase64Decoder",
+    .tp_basicsize = sizeof(StreamingBase64Decoder),
+    .tp_flags = Py_TPFLAGS_DEFAULT,
+    .tp_doc = "StreamingBase64Decoder",
+    .tp_methods = (PyMethodDef[]){
+        {"decode", (PyCFunction)StreamingBase64Decoder_decode, METH_O, ""},
+        {"decode_into", (PyCFunction)(void(*)(void))StreamingBase64Decoder_decode_into, METH_FASTCALL, ""},
+        {"reset", (PyCFunction)StreamingBase64Decoder_reset, METH_NOARGS, ""},
+        {"needs_more_data", (PyCFunction)StreamingBase64Decoder_needs_more_data, METH_NOARGS, ""},
+        {NULL, NULL, 0, NULL},
+    },
+    .tp_new = PyType_GenericNew,
+    .tp_init = StreamingBase64Decoder_init,
+};
+
+static int
+StreamingBase64Encoder_init(PyObject *s, PyObject *args, PyObject *kwds UNUSED) {
+    StreamingBase64Decoder *self = (StreamingBase64Decoder*)s;
+    self->add_trailing_bytes = true;
+    switch (PyTuple_GET_SIZE(args)) {
+        case 0: break;
+        case 1: self->add_trailing_bytes = PyObject_IsTrue(PyTuple_GET_ITEM(args, 0)); break;
+        default: PyErr_SetString(PyExc_TypeError, "constructor takes no more than one argument"); return -1;
+    }
+	base64_stream_encode_init(&self->state, 0);
+    return 0;
+}
+
+static PyObject*
+StreamingBase64Encoder_encode(StreamingBase64Decoder *self, PyObject *a) {
+    RAII_PY_BUFFER(data);
+    if (PyObject_GetBuffer(a, &data, PyBUF_SIMPLE) != 0) return NULL;
+    if (!data.buf || !data.len) return PyBytes_FromStringAndSize(NULL, 0);
+    size_t sz = required_buffer_size_for_base64_encode(data.len);
+    RAII_PyObject(ans, PyBytes_FromStringAndSize(NULL, sz));
+    if (!ans) return NULL;
+    Py_BEGIN_ALLOW_THREADS
+    base64_stream_encode(&self->state, data.buf, data.len, PyBytes_AS_STRING(ans), &sz);
+    Py_END_ALLOW_THREADS
+    if (_PyBytes_Resize(&ans, sz) != 0) return NULL;
+    return Py_NewRef(ans);
+}
+
+static PyObject*
+StreamingBase64Encoder_encode_into(StreamingBase64Decoder *self, PyObject *const *args, Py_ssize_t nargs) {
+    if (nargs != 2) { PyErr_SetString(PyExc_TypeError, "constructor takes exactly two arguments"); return NULL; }
+    RAII_PY_BUFFER(data);
+    if (PyObject_GetBuffer(args[0], &data, PyBUF_WRITE) != 0) return NULL;
+    if (!data.buf || !data.len) return PyLong_FromLong(0);
+    RAII_PY_BUFFER(src);
+    if (PyObject_GetBuffer(args[1], &src, PyBUF_SIMPLE) != 0) return NULL;
+    if (!src.buf || !src.len) return PyLong_FromLong(0);
+    size_t sz = required_buffer_size_for_base64_encode(src.len);
+    if ((Py_ssize_t)sz > data.len) { PyErr_SetString(PyExc_BufferError, "output buffer too small"); return NULL; }
+    Py_BEGIN_ALLOW_THREADS
+    base64_stream_encode(&self->state, src.buf, src.len, data.buf, &sz);
+    Py_END_ALLOW_THREADS
+    return PyLong_FromSize_t(sz);
+}
+
+static PyObject*
+StreamingBase64Encoder_reset(StreamingBase64Decoder *self, PyObject *args UNUSED) {
+    char trailer[4];
+    size_t sz;
+    base64_stream_encode_final(&self->state, trailer, &sz);
+    base64_stream_encode_init(&self->state, 0);
+    if (!self->add_trailing_bytes) { while(sz && trailer[sz-1] == '=') sz--; }
+    return PyBytes_FromStringAndSize(trailer, sz);
+}
+
+static PyTypeObject StreamingBase64Encoder_Type = {
+    PyVarObject_HEAD_INIT(NULL, 0)
+    .tp_name = "kitty.fast_data_types.StreamingBase64Encoder",
+    .tp_basicsize = sizeof(StreamingBase64Decoder),
+    .tp_flags = Py_TPFLAGS_DEFAULT,
+    .tp_doc = "StreamingBase64Encoder",
+    .tp_methods = (PyMethodDef[]){
+        {"encode", (PyCFunction)StreamingBase64Encoder_encode, METH_O, ""},
+        {"encode_into", (PyCFunction)(void(*)(void))StreamingBase64Encoder_encode_into, METH_FASTCALL, ""},
+        {"reset", (PyCFunction)StreamingBase64Encoder_reset, METH_NOARGS, ""},
+        {NULL, NULL, 0, NULL},
+    },
+    .tp_new = PyType_GenericNew,
+    .tp_init = StreamingBase64Encoder_init,
+};
+
 
 static PyObject*
 pyset_iutf8(PyObject UNUSED *self, PyObject *args) {
@@ -223,23 +419,6 @@ locale_is_valid(PyObject *self UNUSED, PyObject *args) {
     Py_RETURN_TRUE;
 }
 
-static PyObject*
-py_getpeereid(PyObject *self UNUSED, PyObject *args) {
-    int fd;
-    if (!PyArg_ParseTuple(args, "i", &fd)) return NULL;
-    uid_t euid = 0; gid_t egid = 0;
-#ifdef __linux__
-    struct ucred cr;
-    socklen_t sz = sizeof(cr);
-    if (getsockopt(fd, SOL_SOCKET, SO_PEERCRED, &cr, &sz) != 0) { PyErr_SetFromErrno(PyExc_OSError); return NULL; }
-    euid = cr.uid; egid = cr.gid;
-#else
-    if (getpeereid(fd, &euid, &egid) != 0) { PyErr_SetFromErrno(PyExc_OSError); return NULL; }
-#endif
-    int u = euid, g = egid;
-    return Py_BuildValue("ii", u, g);
-}
-
 #include "docs_ref_map_generated.h"
 
 static PyObject*
@@ -249,7 +428,7 @@ get_docs_ref_map(PyObject *self UNUSED, PyObject *args UNUSED) {
 
 static PyObject*
 wrapped_kittens(PyObject *self UNUSED, PyObject *args UNUSED) {
-    const char *wrapped_kitten_names = WRAPPED_KITTENS;
+    static const char *wrapped_kitten_names = WRAPPED_KITTENS;
     PyObject *ans = PyUnicode_FromString(wrapped_kitten_names);
     if (ans == NULL) return NULL;
     PyObject *s = PyUnicode_Split(ans, NULL, -1);
@@ -426,13 +605,17 @@ py_timed_debug_print(PyObject *self UNUSED, PyObject *args) {
     Py_RETURN_NONE;
 }
 
+static PyObject*
+py_run_atexit_cleanup_functions(PyObject *self UNUSED, PyObject *args UNUSED) {
+    run_at_exit_cleanup_functions();
+    Py_RETURN_NONE;
+}
 
 static PyMethodDef module_methods[] = {
     METHODB(replace_c0_codes_except_nl_space_tab, METH_O),
     {"wcwidth", (PyCFunction)wcwidth_wrap, METH_O, ""},
     {"expand_ansi_c_escapes", (PyCFunction)expand_ansi_c_escapes, METH_O, ""},
     {"get_docs_ref_map", (PyCFunction)get_docs_ref_map, METH_NOARGS, ""},
-    {"getpeereid", (PyCFunction)py_getpeereid, METH_VARARGS, ""},
     {"wcswidth", (PyCFunction)wcswidth_std, METH_O, ""},
     {"unicode_database_version", (PyCFunction)unicode_database_version, METH_NOARGS, ""},
     {"open_tty", open_tty, METH_VARARGS, ""},
@@ -440,8 +623,10 @@ static PyMethodDef module_methods[] = {
     {"raw_tty", raw_tty, METH_VARARGS, ""},
     {"close_tty", close_tty, METH_VARARGS, ""},
     {"set_iutf8_fd", (PyCFunction)pyset_iutf8, METH_VARARGS, ""},
-    {"base64_encode", (PyCFunction)pybase64_encode, METH_VARARGS, ""},
-    {"base64_decode", (PyCFunction)pybase64_decode, METH_VARARGS, ""},
+    {"base64_encode", (PyCFunction)(void (*) (void))(pybase64_encode), METH_FASTCALL, ""},
+    {"base64_encode_into", (PyCFunction)base64_encode_into, METH_VARARGS, ""},
+    {"base64_decode", (PyCFunction)(void (*) (void))(pybase64_decode), METH_O, ""},
+    {"base64_decode_into", (PyCFunction)base64_decode_into, METH_VARARGS, ""},
     {"thread_write", (PyCFunction)cm_thread_write, METH_VARARGS, ""},
     {"redirect_std_streams", (PyCFunction)redirect_std_streams, METH_VARARGS, ""},
     {"locale_is_valid", (PyCFunction)locale_is_valid, METH_VARARGS, ""},
@@ -452,6 +637,7 @@ static PyMethodDef module_methods[] = {
     {"monotonic", (PyCFunction)py_monotonic, METH_NOARGS, ""},
     {"timed_debug_print", (PyCFunction)py_timed_debug_print, METH_VARARGS, ""},
     {"find_in_memoryview", (PyCFunction)find_in_memoryview, METH_VARARGS, ""},
+    {"run_at_exit_cleanup_functions", (PyCFunction)py_run_atexit_cleanup_functions, METH_NOARGS, ""},
 #ifdef __APPLE__
     METHODB(user_cache_dir, METH_NOARGS),
     METHODB(process_group_map, METH_NOARGS),
@@ -463,13 +649,18 @@ static PyMethodDef module_methods[] = {
     {NULL, NULL, 0, NULL}        /* Sentinel */
 };
 
+static void
+free_fast_data_types_module(void *m UNUSED) {
+    run_at_exit_cleanup_functions();
+}
 
 static struct PyModuleDef module = {
     .m_base = PyModuleDef_HEAD_INIT,
     .m_name = "fast_data_types",   /* name of module */
     .m_doc = NULL,
     .m_size = -1,
-    .m_methods = module_methods
+    .m_methods = module_methods,
+    .m_free = free_fast_data_types_module,
 };
 
 
@@ -483,6 +674,7 @@ extern bool init_child_monitor(PyObject *);
 extern int init_Line(PyObject *);
 extern int init_ColorProfile(PyObject *);
 extern int init_Screen(PyObject *);
+extern bool init_animations(PyObject*);
 extern bool init_fontconfig_library(PyObject*);
 extern bool init_crypto_library(PyObject*);
 extern bool init_desktop(PyObject*);
@@ -499,6 +691,7 @@ extern bool init_logging(PyObject *module);
 extern bool init_png_reader(PyObject *module);
 extern bool init_utmp(PyObject *module);
 extern bool init_loop_utils(PyObject *module);
+extern bool init_systemd_module(PyObject *module);
 #ifdef __APPLE__
 extern int init_CoreText(PyObject *);
 extern bool init_cocoa(PyObject *module);
@@ -523,17 +716,8 @@ shift_to_first_set_bit(CellAttrs x) {
 EXPORTED PyMODINIT_FUNC
 PyInit_fast_data_types(void) {
     PyObject *m;
-    if (sizeof(CellAttrs) != 2u) {
-        PyErr_SetString(PyExc_RuntimeError, "Size of CellAttrs is not 2 on this platform");
-        return NULL;
-    }
-
     m = PyModule_Create(&module);
     if (m == NULL) return NULL;
-    if (Py_AtExit(run_at_exit_cleanup_functions) != 0) {
-        PyErr_SetString(PyExc_RuntimeError, "Failed to register the atexit cleanup handler");
-        return NULL;
-    }
     init_monotonic();
 
     if (!init_logging(m)) return NULL;
@@ -570,6 +754,8 @@ PyInit_fast_data_types(void) {
     if (!init_utmp(m)) return NULL;
     if (!init_loop_utils(m)) return NULL;
     if (!init_crypto_library(m)) return NULL;
+    if (!init_systemd_module(m)) return NULL;
+    if (!init_animations(m)) return NULL;
 
     CellAttrs a;
 #define s(name, attr) { a.val = 0; a.attr = 1; PyModule_AddIntConstant(m, #name, shift_to_first_set_bit(a)); }
@@ -578,7 +764,6 @@ PyInit_fast_data_types(void) {
 #undef s
     PyModule_AddIntConstant(m, "MARK_MASK", MARK_MASK);
     PyModule_AddIntConstant(m, "DECORATION_MASK", DECORATION_MASK);
-    PyModule_AddIntConstant(m, "NUM_UNDERLINE_STYLES", NUM_UNDERLINE_STYLES);
     PyModule_AddStringMacro(m, ERROR_PREFIX);
 #ifdef KITTY_VCS_REV
     PyModule_AddStringMacro(m, KITTY_VCS_REV);
@@ -586,6 +771,7 @@ PyInit_fast_data_types(void) {
     PyModule_AddIntMacro(m, CURSOR_BLOCK);
     PyModule_AddIntMacro(m, CURSOR_BEAM);
     PyModule_AddIntMacro(m, CURSOR_UNDERLINE);
+    PyModule_AddIntMacro(m, CURSOR_HOLLOW);
     PyModule_AddIntMacro(m, NO_CURSOR_SHAPE);
     PyModule_AddIntMacro(m, DECAWM);
     PyModule_AddIntMacro(m, DECCOLM);
@@ -597,6 +783,7 @@ PyInit_fast_data_types(void) {
     PyModule_AddIntMacro(m, ESC_APC);
     PyModule_AddIntMacro(m, ESC_DCS);
     PyModule_AddIntMacro(m, ESC_PM);
+    PyModule_AddIntMacro(m, TEXT_SIZE_CODE);
 #ifdef __APPLE__
     // Apple says its SHM_NAME_MAX but SHM_NAME_MAX is not actually declared in typical CrApple style.
     // This value is based on experimentation and from qsharedmemory.cpp in Qt
@@ -606,5 +793,10 @@ PyInit_fast_data_types(void) {
     PyModule_AddIntConstant(m, "SHM_NAME_MAX", MIN(1023, PATH_MAX));
 #endif
 
+    if (PyType_Ready(&StreamingBase64Decoder_Type) < 0) return NULL;
+    if (PyModule_AddObject(m, "StreamingBase64Decoder", (PyObject *) &StreamingBase64Decoder_Type) < 0) return NULL;
+    if (PyType_Ready(&StreamingBase64Encoder_Type) < 0) return NULL;
+    if (PyModule_AddObject(m, "StreamingBase64Encoder", (PyObject *) &StreamingBase64Encoder_Type) < 0) return NULL;
+
     return m;
 }
