@@ -6,13 +6,11 @@
  */
 
 #include "state.h"
-#include "screen.h"
 #include "charsets.h"
 #include <limits.h>
 #include <math.h>
 #include "glfw-wrapper.h"
 #include "control-codes.h"
-#include "monotonic.h"
 
 extern PyTypeObject Screen_Type;
 
@@ -247,13 +245,13 @@ cell_for_pos(Window *w, unsigned int *x, unsigned int *y, bool *in_left_half_of_
         qx = screen->columns - 1;
         in_left_half = false;
     } else if (mouse_x >= g->left) {
-        double xval = (double)(mouse_x - g->left) / os_window->fonts_data->cell_width;
+        double xval = (double)(mouse_x - g->left) / os_window->fonts_data->fcm.cell_width;
         double fxval = floor(xval);
         qx = (unsigned int)fxval;
         in_left_half = (xval - fxval <= 0.5) ? true : false;
     }
     if (mouse_y >= g->bottom) qy = screen->lines - 1;
-    else if (mouse_y >= g->top) qy = (unsigned int)((double)(mouse_y - g->top) / os_window->fonts_data->cell_height);
+    else if (mouse_y >= g->top) qy = (unsigned int)((double)(mouse_y - g->top) / os_window->fonts_data->fcm.cell_height);
     if (qx < screen->columns && qy < screen->lines) {
         *x = qx; *y = qy;
         *in_left_half_of_cell = in_left_half;
@@ -265,9 +263,11 @@ cell_for_pos(Window *w, unsigned int *x, unsigned int *y, bool *in_left_half_of_
 #define HANDLER(name) static void name(Window UNUSED *w, int UNUSED button, int UNUSED modifiers, unsigned int UNUSED window_idx)
 
 static void
-set_mouse_cursor_when_dragging(void) {
-    if (mouse_cursor_shape != OPT(pointer_shape_when_dragging)) {
-        mouse_cursor_shape = OPT(pointer_shape_when_dragging);
+set_mouse_cursor_when_dragging(Screen *screen) {
+    MouseShape expected_shape = OPT(pointer_shape_when_dragging);
+    if (screen && screen->selections.count && screen->selections.items[0].rectangle_select) expected_shape = OPT(pointer_shape_when_dragging_rectangle);
+    if (mouse_cursor_shape != expected_shape) {
+        mouse_cursor_shape = expected_shape;
         set_mouse_cursor(mouse_cursor_shape);
     }
 }
@@ -278,7 +278,7 @@ update_drag(Window *w) {
     if (screen && screen->selections.in_progress) {
         screen_update_selection(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y, w->mouse_pos.in_left_half_of_cell, (SelectionUpdate){0});
     }
-    set_mouse_cursor_when_dragging();
+    set_mouse_cursor_when_dragging(screen);
 }
 
 static bool
@@ -298,7 +298,7 @@ do_drag_scroll(Window *w, bool upwards) {
 
 bool
 drag_scroll(Window *w, OSWindow *frame) {
-    unsigned int margin = frame->fonts_data->cell_height / 2;
+    unsigned int margin = frame->fonts_data->fcm.cell_height / 2;
     double y = frame->mouse_y;
     bool upwards = y <= (w->geometry.top + margin);
     if (upwards || y >= w->geometry.bottom - margin) {
@@ -360,6 +360,14 @@ detect_url(Screen *screen, unsigned int x, unsigned int y) {
     } else set_mouse_cursor_for_screen(screen);
 }
 
+static bool
+should_handle_in_kitty(Window *w, Screen *screen, int button) {
+    bool in_tracking_mode = (
+        screen->modes.mouse_tracking_mode == ANY_MODE ||
+        (screen->modes.mouse_tracking_mode == MOTION_MODE && button >= 0));
+    return !in_tracking_mode || global_state.active_drag_in_window == w->id;
+}
+
 static bool
 set_mouse_position(Window *w, bool *mouse_cell_changed, bool *cell_half_changed) {
     unsigned int x = 0, y = 0;
@@ -385,11 +393,7 @@ HANDLER(handle_move_event) {
     if (!set_mouse_position(w, &mouse_cell_changed, &cell_half_changed)) return;
     Screen *screen = w->render_data.screen;
     if (OPT(detect_urls)) detect_url(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y);
-    bool in_tracking_mode = (
-        screen->modes.mouse_tracking_mode == ANY_MODE ||
-        (screen->modes.mouse_tracking_mode == MOTION_MODE && button >= 0));
-    bool handle_in_kitty = !in_tracking_mode || global_state.active_drag_in_window == w->id;
-    if (handle_in_kitty) {
+    if (should_handle_in_kitty(w, screen, button)) {
         handle_mouse_movement_in_kitty(w, button, mouse_cell_changed | cell_half_changed);
     } else {
         if (!mouse_cell_changed && screen->modes.mouse_tracking_protocol != SGR_PIXEL_PROTOCOL) return;
@@ -412,7 +416,7 @@ clear_click_queue(Window *w, int button) {
 
 static double
 radius_for_multiclick(void) {
-    return 0.5 * (global_state.callback_os_window ? global_state.callback_os_window->fonts_data->cell_height : 8);
+    return 0.5 * (global_state.callback_os_window ? global_state.callback_os_window->fonts_data->fcm.cell_height : 8);
 }
 
 static bool
@@ -487,42 +491,50 @@ move_cursor_to_mouse_if_at_shell_prompt(Window *w) {
     Screen *screen = w->render_data.screen;
     int y = screen_cursor_at_a_shell_prompt(screen);
     if (y < 0 || (unsigned)y > w->mouse_pos.cell_y) return false;
-    return screen_fake_move_cursor_to_position(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y);
+
+    if (screen_prompt_supports_click_events(screen)) {
+        int sz = encode_mouse_event_impl(&w->mouse_pos, SGR_PROTOCOL, 1, PRESS, 0);
+        if (sz > 0) {
+            mouse_event_buf[sz] = 0;
+            write_escape_code_to_child(screen, ESC_CSI, mouse_event_buf);
+            return true;
+        }
+
+        return false;
+    } else {
+        return screen_fake_move_cursor_to_position(screen, w->mouse_pos.cell_x, w->mouse_pos.cell_y);
+    }
 }
 
 
-typedef struct PendingClick {
-    id_type window_id;
-    int button, count, modifiers;
-    bool grabbed;
-    monotonic_t at;
-    MousePosition mouse_pos;
-    unsigned long press_num;
-    double radius_for_multiclick;
-} PendingClick;
-
-static void
-free_pending_click(id_type timer_id UNUSED, void *pc) { free(pc); }
-
 void
-send_pending_click_to_window(Window *w, void *data) {
-    PendingClick *pc = (PendingClick*)data;
-    const ClickQueue *q = &w->click_queues[pc->button];
+send_pending_click_to_window(Window *w, int i) {
+    const id_type wid = w->id;
+    if (i < 0) {
+        while(true) {
+            w = window_for_id(wid);
+            if (!w || !w->pending_clicks.num) break;
+            send_pending_click_to_window(w, w->pending_clicks.num - 1);
+        }
+        return;
+    }
+    PendingClick pc = w->pending_clicks.clicks[i];
+    remove_i_from_array(w->pending_clicks.clicks, (unsigned)i, w->pending_clicks.num);
+    const ClickQueue *q = &w->click_queues[pc.button];
     // only send click if no presses have happened since the release that triggered
     // the click or if the subsequent press is too far or too late for a double click
     if (!q->length) return;
 #define press(n) q->clicks[q->length - n]
     if (
-            press(1).at <= pc->at || // latest press is before click release
-            (q->length > 1 && press(2).num == pc->press_num &&  (   // penultimate press is the press that belongs to this click
+            press(1).at <= pc.at || // latest press is before click release
+            (q->length > 1 && press(2).num == pc.press_num &&  (   // penultimate press is the press that belongs to this click
                 press(1).at - press(2).at > OPT(click_interval) ||  // too long between the presses for it to be a double click
-                distance(press(1).x, press(1).y, press(2).x, press(2).y) > pc->radius_for_multiclick  // presses are too far apart
+                distance(press(1).x, press(1).y, press(2).x, press(2).y) > pc.radius_for_multiclick  // presses are too far apart
             ))
     ) {
         MousePosition current_pos = w->mouse_pos;
-        w->mouse_pos = pc->mouse_pos;
-        id_type wid = w->id;
-        dispatch_mouse_event(w, pc->button, pc->count, pc->modifiers, pc->grabbed);
+        w->mouse_pos = pc.mouse_pos;
+        dispatch_mouse_event(w, pc.button, pc.count, pc.modifiers, pc.grabbed);
         w = window_for_id(wid);
         if (w) w->mouse_pos = current_pos;
     }
@@ -534,20 +546,20 @@ dispatch_possible_click(Window *w, int button, int modifiers) {
     Screen *screen = w->render_data.screen;
     int count = multi_click_count(w, button);
     if (release_is_click(w, button)) {
-        PendingClick *pc = calloc(1, sizeof(PendingClick));
-        if (pc) {
-            const ClickQueue *q = &w->click_queues[button];
-            pc->press_num = q->length ? q->clicks[q->length - 1].num : 0;
-            pc->window_id = w->id;
-            pc->mouse_pos = w->mouse_pos;
-            pc->at = monotonic();
-            pc->button = button;
-            pc->count = count == 2 ? -3 : -2;
-            pc->modifiers = modifiers;
-            pc->grabbed = screen->modes.mouse_tracking_mode != 0;
-            pc->radius_for_multiclick = radius_for_multiclick();
-            add_main_loop_timer(OPT(click_interval), false, send_pending_click_to_window_id, pc, free_pending_click);
-        }
+        ensure_space_for(&(w->pending_clicks), clicks, PendingClick, w->pending_clicks.num + 1, capacity, 4, true);
+        PendingClick *pc = w->pending_clicks.clicks + w->pending_clicks.num++;
+        zero_at_ptr(pc);
+        const ClickQueue *q = &w->click_queues[button];
+        pc->press_num = q->length ? q->clicks[q->length - 1].num : 0;
+        pc->window_id = w->id;
+        pc->mouse_pos = w->mouse_pos;
+        pc->at = monotonic();
+        pc->button = button;
+        pc->count = count == 2 ? -3 : -2;
+        pc->modifiers = modifiers;
+        pc->grabbed = screen->modes.mouse_tracking_mode != 0;
+        pc->radius_for_multiclick = radius_for_multiclick();
+        add_main_loop_timer(OPT(click_interval), false, dispatch_pending_clicks, NULL, NULL);
     }
 }
 
@@ -683,7 +695,7 @@ update_mouse_pointer_shape(void) {
 }
 
 void
-enter_event(void) {
+enter_event(int modifiers) {
 #ifdef __APPLE__
     // On cocoa there is no way to configure the window manager to
     // focus windows on mouse enter, so we do it ourselves
@@ -691,6 +703,20 @@ enter_event(void) {
         focus_os_window(global_state.callback_os_window, false, NULL);
     }
 #endif
+    // If the mouse is grabbed send a move event to update the cursor position
+    // since the last report.
+    if (global_state.redirect_mouse_handling || global_state.active_drag_in_window || global_state.tracked_drag_in_window) return;
+    unsigned window_idx; bool in_tab_bar;
+    Window *w = window_for_event(&window_idx, &in_tab_bar);
+    if (!w || in_tab_bar) return;
+    bool mouse_cell_changed = false;
+    bool cell_half_changed = false;
+    if (!set_mouse_position(w, &mouse_cell_changed, &cell_half_changed)) return;
+    Screen *screen = w->render_data.screen;
+    int button = currently_pressed_button();
+    if (!screen || should_handle_in_kitty(w, screen, button)) return;
+    int sz = encode_mouse_button(w, button, button >=0 ? DRAG : MOVE, modifiers);
+    if (sz > 0) { mouse_event_buf[sz] = 0; write_escape_code_to_child(screen, ESC_CSI, mouse_event_buf); }
 }
 
 static void
@@ -753,7 +779,7 @@ mouse_selection(Window *w, int code, int button) {
             extend_selection(w, false, false);
             break;
     }
-    set_mouse_cursor_when_dragging();
+    set_mouse_cursor_when_dragging(screen);
 #undef S
 }
 
@@ -949,7 +975,7 @@ scroll_event(double xoffset, double yoffset, int flags, int modifiers) {
     bool is_high_resolution = flags & 1;
 
     if (yoffset != 0.0) {
-        s = scale_scroll(screen->modes.mouse_tracking_mode, yoffset, is_high_resolution, &screen->pending_scroll_pixels_y, global_state.callback_os_window->fonts_data->cell_height);
+        s = scale_scroll(screen->modes.mouse_tracking_mode, yoffset, is_high_resolution, &screen->pending_scroll_pixels_y, global_state.callback_os_window->fonts_data->fcm.cell_height);
         if (s) {
             bool upwards = s > 0;
             if (screen->modes.mouse_tracking_mode) {
@@ -961,13 +987,16 @@ scroll_event(double xoffset, double yoffset, int flags, int modifiers) {
                     }
                 }
             } else {
-                if (screen->linebuf == screen->main_linebuf) screen_history_scroll(screen, abs(s), upwards);
+                if (screen->linebuf == screen->main_linebuf) {
+                    screen_history_scroll(screen, abs(s), upwards);
+                    if (screen->selections.in_progress) update_drag(w);
+                }
                 else fake_scroll(w, abs(s), upwards);
             }
         }
     }
     if (xoffset != 0.0) {
-        s = scale_scroll(screen->modes.mouse_tracking_mode, xoffset, is_high_resolution, &screen->pending_scroll_pixels_x, global_state.callback_os_window->fonts_data->cell_width);
+        s = scale_scroll(screen->modes.mouse_tracking_mode, xoffset, is_high_resolution, &screen->pending_scroll_pixels_x, global_state.callback_os_window->fonts_data->fcm.cell_width);
         if (s) {
             if (screen->modes.mouse_tracking_mode) {
                 int sz = encode_mouse_scroll(w, s > 0 ? 6 : 7, modifiers);
