@@ -1,27 +1,21 @@
-#!/usr/bin/env python3
+#!/usr/bin/env python
 # License: GPL v3 Copyright: 2018, Kovid Goyal <kovid at kovidgoyal.net>
 
 import os
 import re
+from collections.abc import Callable, Sequence
 from functools import lru_cache, partial, wraps
 from string import Formatter as StringFormatter
 from typing import (
     Any,
-    Callable,
-    Dict,
-    List,
     NamedTuple,
-    Optional,
-    Sequence,
-    Tuple,
-    Union,
 )
 
 from .borders import Border, BorderColor
-from .config import build_ansi_color_table
 from .constants import config_dir
 from .fast_data_types import (
     DECAWM,
+    TOP_EDGE,
     Color,
     Region,
     Screen,
@@ -32,7 +26,9 @@ from .fast_data_types import (
     set_tab_bar_render_data,
     update_tab_bar_edge_colors,
     viewport_for_window,
+    wcswidth,
 )
+from .progress import ProgressState
 from .rgb import alpha_blend, color_as_sgr, color_from_int, to_color
 from .types import WindowGeometry, run_once
 from .typing import EdgeLiteral, PowerlineStyle
@@ -48,10 +44,13 @@ class TabBarData(NamedTuple):
     num_window_groups: int
     layout_name: str
     has_activity_since_last_focus: bool
-    active_fg: Optional[int]
-    active_bg: Optional[int]
-    inactive_fg: Optional[int]
-    inactive_bg: Optional[int]
+    active_fg: int | None
+    active_bg: int | None
+    inactive_fg: int | None
+    inactive_bg: int | None
+    num_of_windows_with_progress: int
+    total_progress: int
+    last_focused_window_with_progress_id: int
 
 
 class DrawData(NamedTuple):
@@ -66,7 +65,7 @@ class DrawData(NamedTuple):
     inactive_bg: Color
     default_bg: Color
     title_template: str
-    active_title_template: Optional[str]
+    active_title_template: str | None
     tab_activity_symbol: str
     powerline_style: PowerlineStyle
     tab_bar_edge: EdgeLiteral
@@ -95,12 +94,12 @@ def as_rgb(x: int) -> int:
     return (x << 8) | 2
 
 
-@lru_cache()
+@lru_cache
 def report_template_failure(template: str, e: str) -> None:
     log_error(f'Invalid tab title template: "{template}" with error: {e}')
 
 
-@lru_cache()
+@lru_cache
 def compile_template(template: str) -> Any:
     try:
         return compile('f"""' + template + '"""', '<template>', 'eval')
@@ -123,6 +122,8 @@ class ColorFormatter:
         elif q == 'tab':
             col = color_from_int((self.draw_data.tab_bg if self.which == '4' else self.draw_data.tab_fg)(self.tab_data))
             ans = f'8{color_as_sgr(col)}'
+        elif q.startswith('color'):
+            ans = f'8:5:{int(q[5:])}'
         else:
             if name.startswith('_'):
                 q = f'#{name[1:]}'
@@ -144,7 +145,7 @@ class Formatter:
 
 
 @run_once
-def super_sub_maps() -> Tuple[Dict[int, Union[None, int]], Dict[int, Union[None, int]]]:
+def super_sub_maps() -> tuple[dict[int, int], dict[int, int]]:
     import string
     sup_table = str.maketrans(
         string.ascii_lowercase + string.ascii_uppercase + string.digits + '+-=()',
@@ -157,7 +158,7 @@ def super_sub_maps() -> Tuple[Dict[int, Union[None, int]], Dict[int, Union[None,
 
 class SupSub:
 
-    def __init__(self, data: Dict[str, Any], is_subscript: bool = False):
+    def __init__(self, data: dict[str, Any], is_subscript: bool = False):
         self.__data = data
         self.__is_subscript = is_subscript
 
@@ -168,8 +169,8 @@ class SupSub:
 
 
 class ExtraData:
-    prev_tab: Optional[TabBarData] = None
-    next_tab: Optional[TabBarData] = None
+    prev_tab: TabBarData | None = None
+    next_tab: TabBarData | None = None
     # true if the draw_tab function is called just for layout. In such cases,
     # if drawing is expensive the draw_tab function should avoid drawing and
     # just move the cursor to its final position, as if drawing was performed.
@@ -221,9 +222,40 @@ class TabAccessor:
         tab = get_boss().tab_for_id(self.tab_id)
         return os.path.basename((tab.get_exe_of_active_window(oldest=True) if tab else '') or '')
 
+    @property
+    def last_focused_progress_percent(self) -> str:
+        tab = get_boss().tab_for_id(self.tab_id)
+        if not tab or not tab.last_focused_window_with_progress_id:
+            return ''
+        w = get_boss().window_id_map.get(tab.last_focused_window_with_progress_id)
+        if w is None:
+            return ''
+        if w.progress.state is ProgressState.error:
+            return '  '
+        if w.progress.state is ProgressState.unset:
+            return ''
+        if w.progress.state is ProgressState.indeterminate:
+            return '  '
+        p = f'{w.progress.percent}% '
+        if w.progress.state is ProgressState.paused:
+            return f'  {p}'
+        return p
+
+    @property
+    def progress_percent(self) -> str:
+        tab = get_boss().tab_for_id(self.tab_id)
+        if not tab or not tab.last_focused_window_with_progress_id:
+            return ''
+        if tab.num_of_windows_with_progress <= 1:
+            return self.last_focused_progress_percent
+        p = int(tab.total_progress / tab.num_of_windows_with_progress)
+        return f'{p}% '
+
+
 
 safe_builtins = {
-    'max': max, 'min': min, 'str': str, 'repr': repr, 'abs': abs, 'len': len, 're': re,
+    'max': max, 'min': min, 'str': str, 'repr': repr, 'abs': abs, 'len': len, 'chr': chr, 'ord': ord, 're': re,
+    'wcswidth': wcswidth,
 }
 
 
@@ -241,6 +273,7 @@ def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int,
         max_title_length = min(max_title_length, draw_data.max_tab_title_length)
     ColorFormatter.draw_data = draw_data
     ColorFormatter.tab_data = tab
+    boss = get_boss()
     eval_locals = {
         'index': index,
         'layout_name': tab.layout_name,
@@ -254,6 +287,7 @@ def draw_title(draw_data: DrawData, screen: Screen, tab: TabBarData, index: int,
         'bell_symbol': draw_data.bell_on_tab if tab.needs_attention else '',
         'activity_symbol': draw_data.tab_activity_symbol if tab.has_activity_since_last_focus else '',
         'max_title_length': max_title_length,
+        'keyboard_mode': boss.mappings.current_keyboard_mode_name,
     }
     template = draw_data.title_template
     if tab.is_active and draw_data.active_title_template is not None:
@@ -384,7 +418,7 @@ def draw_tab_with_fade(
     return end
 
 
-powerline_symbols: Dict[PowerlineStyle, Tuple[str, str]] = {
+powerline_symbols: dict[PowerlineStyle, tuple[str, str]] = {
     'slanted': ('', '╱'),
     'round': ('', '')
 }
@@ -480,8 +514,8 @@ class TabBar:
         self.os_window_id = os_window_id
         self.num_tabs = 1
         self.data_buffer_size = 0
-        self.blank_rects: Tuple[Border, ...] = ()
-        self.cell_ranges: List[Tuple[int, int]] = []
+        self.blank_rects: tuple[Border, ...] = ()
+        self.cell_ranges: list[tuple[int, int]] = []
         self.laid_out_once = False
         self.apply_options()
 
@@ -494,11 +528,8 @@ class TabBar:
             self.screen = s = Screen(None, 1, 10, 0, self.cell_width, cell_height)
         else:
             s = self.screen
-        s.color_profile.update_ansi_color_table(build_ansi_color_table(opts))
-        s.color_profile.set_configured_colors(
-            color_as_int(opts.inactive_tab_foreground),
-            color_as_int(opts.tab_bar_background or opts.background)
-        )
+        s.color_profile.default_fg = opts.inactive_tab_foreground
+        s.color_profile.default_bg = opts.tab_bar_background or opts.background
         sep = opts.tab_separator
         self.trailing_spaces = self.leading_spaces = 0
         while sep and sep[0] == ' ':
@@ -521,7 +552,7 @@ class TabBar:
             opts.active_tab_title_template,
             opts.tab_activity_symbol,
             opts.tab_powerline_style,
-            'top' if opts.tab_bar_edge == 1 else 'bottom',
+            'top' if opts.tab_bar_edge == TOP_EDGE else 'bottom',
             opts.tab_title_max_length,
         )
         ts = opts.tab_bar_style
@@ -542,7 +573,7 @@ class TabBar:
         else:
             self.align = lambda: None
 
-    def patch_colors(self, spec: Dict[str, Optional[int]]) -> None:
+    def patch_colors(self, spec: dict[str, int | None]) -> None:
         opts = get_options()
         atf = spec.get('active_tab_foreground')
         if isinstance(atf, int):
@@ -574,10 +605,12 @@ class TabBar:
             ifg = color_from_int(fg)
             if ifg is not None:
                 self.draw_data = self.draw_data._replace(inactive_fg=ifg)
-        self.screen.color_profile.set_configured_colors(fg, bg)
+        self.screen.color_profile.reload_from_opts()
+        self.screen.color_profile.default_fg = color_from_int(fg)
+        self.screen.color_profile.default_bg = color_from_int(bg)
 
     @property
-    def current_colors(self) -> Dict[str, Color]:
+    def current_colors(self) -> dict[str, Color]:
         return {
             'active_tab_foreground': self.draw_data.active_fg,
             'inactive_tab_foreground': self.draw_data.inactive_fg,
@@ -588,7 +621,7 @@ class TabBar:
 
     def update_blank_rects(self, central: Region, tab_bar: Region, vw: int, vh: int) -> None:
         opts = get_options()
-        blank_rects: List[Border] = []
+        blank_rects: list[Border] = []
         bg = BorderColor.tab_bar_margin_color if opts.tab_bar_margin_color is not None else BorderColor.default_bg
         if opts.tab_bar_margin_height:
             if opts.tab_bar_edge == 3:  # bottom
@@ -603,7 +636,7 @@ class TabBar:
                     blank_rects.append(Border(0, tab_bar.bottom + 1, vw, central.top, bg))
         g = self.window_geometry
         left_bg = right_bg = bg
-        if opts.tab_bar_margin_color is None:
+        if opts.tab_bar_margin_color is None or opts.tab_bar_margin_width == 0:
             left_bg = BorderColor.tab_bar_left_edge_color
             right_bg = BorderColor.tab_bar_right_edge_color
         if g.left > 0:
@@ -636,7 +669,7 @@ class TabBar:
         last_tab = data[-1] if data else None
         ed = ExtraData()
 
-        def draw_tab(i: int, tab: TabBarData, cell_ranges: List[Tuple[int, int]], max_tab_length: int) -> None:
+        def draw_tab(i: int, tab: TabBarData, cell_ranges: list[tuple[int, int]], max_tab_length: int) -> None:
             ed.prev_tab = data[i - 1] if i > 0 else None
             ed.next_tab = data[i + 1] if i + 1 < len(data) else None
             s.cursor.bg = as_rgb(self.draw_data.tab_bg(t))
@@ -685,7 +718,7 @@ class TabBar:
 
         s.cursor.x = 0
         s.erase_in_line(2, False)
-        cr: List[Tuple[int, int]] = []
+        cr: list[tuple[int, int]] = []
         ed.for_layout = False
         for i, t in enumerate(data):
             try:
@@ -711,7 +744,7 @@ class TabBar:
         self.screen.reset_callbacks()
         del self.screen
 
-    def tab_at(self, x: int) -> Optional[int]:
+    def tab_at(self, x: int) -> int | None:
         if self.laid_out_once:
             x = (x - self.window_geometry.left) // self.cell_width
             for i, (a, b) in enumerate(self.cell_ranges):
