@@ -15,13 +15,13 @@ import time
 from contextlib import contextmanager, suppress
 from functools import wraps
 from pty import CHILD, STDIN_FILENO, STDOUT_FILENO, fork
-from typing import Optional
 from unittest import TestCase
 
 from kitty.config import finalize_keys, finalize_mouse_mappings
-from kitty.fast_data_types import Cursor, HistoryBuf, LineBuf, Screen, get_options, monotonic, set_options
+from kitty.fast_data_types import TEXT_SIZE_CODE, Cursor, HistoryBuf, LineBuf, Screen, get_options, monotonic, set_options
 from kitty.options.parse import merge_result_dicts
 from kitty.options.types import Options, defaults
+from kitty.rgb import to_color
 from kitty.types import MouseEvent
 from kitty.utils import read_screen_size
 from kitty.window import decode_cmdline, process_remote_print, process_title_from_child
@@ -36,6 +36,13 @@ def parse_bytes(screen, data, dump_callback=None):
         screen.test_parse_written_data(dump_callback)
 
 
+def draw_multicell(
+    screen: Screen, text: str, width: int = 0, scale: int = 1, subscale_n: int = 0, subscale_d: int = 0, vertical_align: int = 0, horizontal_align: int = 0
+    ) -> None:
+    cmd = f'\x1b]{TEXT_SIZE_CODE};w={width}:s={scale}:n={subscale_n}:d={subscale_d}:v={vertical_align}:h={horizontal_align};{text}\a'
+    parse_bytes(screen, cmd.encode())
+
+
 class Callbacks:
 
     def __init__(self, pty=None) -> None:
@@ -50,6 +57,23 @@ class Callbacks:
     def write(self, data) -> None:
         self.wtcbuf += bytes(data)
 
+    def notify_child_of_resize(self):
+        self.num_of_resize_events += 1
+
+    def color_control(self, code, data) -> None:
+        from kitty.window import color_control
+        response = color_control(self.color_profile, code, data)
+        if response:
+            def p(x):
+                if '@' in x:
+                    return (to_color(x.partition('@')[0]), int(255 * float(x.partition('@')[2])))
+                ans = to_color(x)
+                if ans is None:
+                    ans = x
+                return ans
+            parts = {x.partition('=')[0]:p(x.partition('=')[2]) for x in response.split(';')[1:]}
+            self.color_control_responses.append(parts)
+
     def title_changed(self, data, is_base64=False) -> None:
         self.titlebuf.append(process_title_from_child(data, is_base64, ''))
 
@@ -68,7 +92,7 @@ class Callbacks:
     def color_profile_popped(self, x) -> None:
         pass
 
-    def cmd_output_marking(self, is_start: Optional[bool], data: str = '') -> None:
+    def cmd_output_marking(self, is_start: bool | None, data: str = '') -> None:
         if is_start:
             self.last_cmd_at = monotonic()
             self.last_cmd_cmdline = decode_cmdline(data) if data else data
@@ -97,6 +121,7 @@ class Callbacks:
         self.iconbuf = self.colorbuf = self.ctbuf = ''
         self.titlebuf = []
         self.printbuf = []
+        self.color_control_responses = []
         self.notifications = []
         self.open_urls = []
         self.cc_buf = []
@@ -106,6 +131,7 @@ class Callbacks:
         self.last_cmd_exit_status = sys.maxsize
         self.last_cmd_cmdline = ''
         self.last_cmd_at = 0
+        self.num_of_resize_events = 0
 
     def on_bell(self) -> None:
         self.bell_count += 1
@@ -238,6 +264,7 @@ class BaseTest(TestCase):
         self.set_options(options)
         c = Callbacks()
         s = Screen(c, lines, cols, scrollback, cell_width, cell_height, 0, c)
+        c.color_profile = s.color_profile
         return s
 
     def create_pty(
@@ -287,6 +314,7 @@ class PTY:
         if argv is None:
             from kitty.child import openpty
             self.master_fd, self.slave_fd = openpty()
+            self.child_pid = 0
         else:
             self.child_pid, self.master_fd = fork()
             self.is_child = self.child_pid == CHILD
@@ -369,7 +397,12 @@ class PTY:
     def wait_till(self, q, timeout=10, timeout_msg=None):
         end_time = time.monotonic() + timeout
         while not q() and time.monotonic() <= end_time:
-            self.process_input_from_child(timeout=end_time - time.monotonic())
+            try:
+                self.process_input_from_child(timeout=end_time - time.monotonic())
+            except OSError as e:
+                if not q():
+                    raise Exception(f'Failed to read from pty with error: {e}. Screen contents:  \n {repr(self.screen_contents())}') from e
+                return
         if not q():
             msg = 'The condition was not met'
             if timeout_msg is not None:
