@@ -5,10 +5,8 @@ import os
 import sys
 import tempfile
 
-from kitty.config import build_ansi_color_table, defaults
 from kitty.fast_data_types import (
     Color,
-    ColorProfile,
     HistoryBuf,
     LineBuf,
     expand_ansi_c_escapes,
@@ -206,9 +204,9 @@ class TestDataTypes(BaseTest):
         l0.add_combining_char(0, '\u0302')
         self.ae(l0[0], ' \u0300\U000e0100\u0302')
         l0.add_combining_char(0, '\u0301')
-        self.ae(l0[0], ' \u0300\U000e0100\u0301')
+        self.ae(l0[0], ' \u0300\U000e0100\u0302\u0301')
         self.ae(l0[1], '\0')
-        self.ae(str(l0), ' \u0300\U000e0100\u0301')
+        self.ae(str(l0), ' \u0300\U000e0100\u0302\u0301')
         t = 'Testing with simple text'
         lb = LineBuf(2, len(t))
         l0 = lb.line(0)
@@ -242,19 +240,6 @@ class TestDataTypes(BaseTest):
         l3 = lb.line(0)
         l3.set_text(t, 0, len(t), C())
         self.ae(t, str(l3))
-        l3.right_shift(4, 2)
-        self.ae('0123454567', str(l3))
-        l3.set_text(t, 0, len(t), C())
-        l3.right_shift(0, 0)
-        self.ae(t, str(l3))
-        l3.right_shift(0, 1)
-        self.ae(str(l3), '0' + t[:-1])
-        l3.set_text(t, 0, len(t), C())
-        l3.left_shift(0, 2)
-        self.ae(str(l3), t[2:] + '89')
-        l3.set_text(t, 0, len(t), C())
-        l3.left_shift(7, 3)
-        self.ae(str(l3), t)
 
         l3.set_text(t, 0, len(t), C())
         q = C()
@@ -329,28 +314,25 @@ class TestDataTypes(BaseTest):
         self.ae(l4.url_end_at(0), len(l4) - 2)
         self.ae(l4.url_end_at(0, 0, True), len(l4) - 1)
 
-    def rewrap(self, lb, lb2):
-        hb = HistoryBuf(lb2.ynum, lb2.xnum)
-        cy = lb.rewrap(lb2, hb)
-        return hb, cy[1]
+    def rewrap(self, lb, lines, columns):
+        return lb.rewrap(lines, columns)
 
     def test_rewrap_simple(self):
         ' Same width buffers '
         lb = filled_line_buf(5, 5)
         lb2 = LineBuf(lb.ynum, lb.xnum)
-        self.rewrap(lb, lb2)
+        lb2 = self.rewrap(lb, lb.ynum, lb.xnum)[0]
         for i in range(lb.ynum):
             self.ae(lb2.line(i), lb.line(i))
-        lb2 = LineBuf(8, 5)
-        cy = self.rewrap(lb, lb2)[1]
+        lb2, _, cy = self.rewrap(lb, 8, 5)
         self.ae(cy, 5)
         for i in range(lb.ynum):
-            self.ae(lb2.line(i), lb.line(i))
+            self.ae(lb2.line(i), lb.line(i), i)
         empty = LineBuf(1, lb2.xnum)
         for i in range(lb.ynum, lb2.ynum):
             self.ae(str(lb2.line(i)), str(empty.line(0)))
         lb2 = LineBuf(3, 5)
-        cy = self.rewrap(lb, lb2)[1]
+        lb2, _, cy = self.rewrap(lb, 3, 5)
         self.ae(cy, 3)
         for i in range(lb2.ynum):
             self.ae(lb2.line(i), lb.line(i + 2))
@@ -363,8 +345,7 @@ class TestDataTypes(BaseTest):
             self.ae(l0, str(l2))
 
     def line_comparison_rewrap(self, lb, *lines):
-        lb2 = LineBuf(len(lines), max(map(len, lines)))
-        self.rewrap(lb, lb2)
+        lb2 = self.rewrap(lb, len(lines), max(map(len, lines)))[0]
         self.line_comparison(lb2, *lines)
         return lb2
 
@@ -476,14 +457,6 @@ class TestDataTypes(BaseTest):
         self.ae(sanitize_url_for_dispay_to_user(
             'h://a\u0430b.com/El%20Ni%C3%B1o/'), 'h://xn--ab-7kc.com/El Niño/')
 
-    def test_color_profile(self):
-        c = ColorProfile()
-        c.update_ansi_color_table(build_ansi_color_table())
-        for i in range(8):
-            col = getattr(defaults, f'color{i}')
-            self.ae(c.as_color(i << 8 | 1), col)
-        self.ae(c.as_color(255 << 8 | 1), Color(0xee, 0xee, 0xee))
-
     def test_historybuf(self):
         lb = filled_line_buf()
         hb = HistoryBuf(5, 5)
@@ -509,35 +482,31 @@ class TestDataTypes(BaseTest):
             self.ae(str(hb.line(i)).rstrip(), str(3000 - 1 - i))
 
         # rewrap
+        def as_ansi(hb):
+            lines = []
+            hb.as_ansi(lines.append)
+            return ''.join(lines)
         hb = filled_history_buf(5, 5)
-        hb2 = HistoryBuf(hb.ynum, hb.xnum)
-        hb.rewrap(hb2)
+        for i in range(hb.ynum):
+            hb.line(i).set_wrapped_flag(True)
+        before = as_ansi(hb)
+        hb2 = hb.rewrap(10)
+        self.ae(before, as_ansi(hb2).rstrip())
+
+        hb = filled_history_buf(5, 5)
+        hb2 = hb.rewrap(hb.xnum)
         for i in range(hb.ynum):
             self.ae(hb2.line(i), hb.line(i))
-        hb2 = HistoryBuf(8, 5)
-        hb.rewrap(hb2)
-        for i in range(hb.ynum):
-            self.ae(hb2.line(i), hb.line(i))
-        for i in range(hb.ynum, hb2.ynum):
-            with self.assertRaises(IndexError):
-                hb2.line(i)
-        hb2 = HistoryBuf(3, 5)
-        hb.rewrap(hb2)
-        for i in range(hb2.ynum):
-            self.ae(hb2.line(i), hb.line(i))
-        self.ae(hb2.dirty_lines(), list(range(hb2.ynum)))
         hb = filled_history_buf(5, 5)
-        hb2 = HistoryBuf(hb.ynum, hb.xnum * 2)
-        hb.rewrap(hb2)
+        hb2 = hb.rewrap(hb.xnum * 2)
         hb3 = HistoryBuf(hb.ynum, hb.xnum)
-        hb2.rewrap(hb3)
+        hb3 = hb2.rewrap(hb.xnum)
         for i in range(hb.ynum):
             self.ae(hb.line(i), hb3.line(i))
 
         hb2 = HistoryBuf(hb.ynum, hb.xnum)
-        large_hb.rewrap(hb2)
-        hb2 = HistoryBuf(large_hb.ynum, large_hb.xnum)
-        large_hb.rewrap(hb2)
+        hb2 = large_hb.rewrap(hb.xnum)
+        hb2.rewrap(large_hb.xnum)
 
     def test_ansi_repr(self):
         lb = filled_line_buf()
@@ -553,8 +522,8 @@ class TestDataTypes(BaseTest):
         c.bg = (1 << 24) | (2 << 16) | (3 << 8) | 2
         c.decoration_fg = (5 << 8) | 1
         l2.set_text('1', 0, 1, c)
-        self.ae(l2.as_ansi(), '\x1b[1;2;3;7;9;34;48:2:1:2:3;58:5:5m' '1'
-                '\x1b[22;23;27;29;39;49;59m' '0000')
+        self.ae(str(l2), '10000')
+        self.ae(l2.as_ansi(), '\x1b[1;2;3;7;9;34;48:2:1:2:3;58:5:5m' '1' '\x1b[22;23;27;29;39;49;59m' '0000')  # ]]
         lb = filled_line_buf()
         for i in range(1, lb.ynum + 1):
             lb.set_continued(i, True)
@@ -570,9 +539,10 @@ class TestDataTypes(BaseTest):
         def q(x, y=''):
             self.ae(y or x, strip_csi(x))
         q('test')
-        q('a\x1bbc', 'ac')
+        q('a\x1bbc', 'abc')
         q('a\x1b[bc', 'ac')
         q('a\x1b[12;34:43mbc', 'abc')
+        q('a\x1b[12;34:43\U0001f638', 'a\U0001f638')
 
     def test_single_key(self):
         from kitty.fast_data_types import GLFW_MOD_KITTY, GLFW_MOD_SHIFT, SingleKey
@@ -594,7 +564,7 @@ class TestDataTypes(BaseTest):
         self.assertNotEqual(SingleKey(key=1, mods=2), SingleKey(key=1))
 
     def test_notify_identifier_sanitization(self):
-        from kitty.notify import sanitize_identifier_pat
+        from kitty.notifications import sanitize_identifier_pat
         self.ae(sanitize_identifier_pat().sub('', '\x1b\nabc\n[*'), 'abc')
 
     def test_bracketed_paste_sanitizer(self):
@@ -602,7 +572,7 @@ class TestDataTypes(BaseTest):
         for x in ('\x1b[201~ab\x9b201~cd', '\x1b[201\x1b[201~~ab'):  # ]]]
             q = sanitize_for_bracketed_paste(x.encode('utf-8'))
             self.assertNotIn(b'\x1b[201~', q)
-            self.assertNotIn('\x9b201~'.encode('utf-8'), q)
+            self.assertNotIn('\x9b201~'.encode(), q)
             self.assertIn(b'ab', q)
 
     def test_expand_ansi_c_escapes(self):
@@ -629,7 +599,7 @@ class TestDataTypes(BaseTest):
 
     def test_shlex_split(self):
         for bad in (
-            'abc\\', '\\', "'abc", "'", '"', 'asd' + '\\',
+            'abc\\', '\\', "'abc", "'", '"', 'asd' + '\\', r'"a\"', '"a\\',
         ):
             with self.assertRaises(ValueError, msg=f'Failed to raise exception for {bad!r}'):
                 tuple(shlex_split_with_positions(bad))
@@ -639,7 +609,25 @@ class TestDataTypes(BaseTest):
             r'x "ab"y \m': ((0, 'x'), (2, 'aby'), (8, 'm')),
             r'''x'y"\z'1''': ((0, 'xy"\\z1'),),
             r'\abc\ d': ((0, 'abc d'),),
-            '': (), '   ': (), ' \tabc\n\t\r ': ((2, 'abc'),),
+            '': ((0, ''),), '   ': ((0, ''),), ' \tabc\n\t\r ': ((2, 'abc'),),
+            "$'ab'": ((0, '$ab'),),
         }.items():
             actual = tuple(shlex_split_with_positions(q))
             self.ae(expected, actual, f'Failed for text: {q!r}')
+
+        for q, expected in {
+            "$'ab'": ((0, 'ab'),),
+            "1$'ab'": ((0, '1ab'),),
+            '''"1$'ab'"''': ((0, "1$'ab'"),),
+            r"$'a\123b'": ((0, 'a\123b'),),
+            r"$'a\1b'": ((0, 'a\001b'),),
+            r"$'a\12b'": ((0, 'a\012b'),),
+            r"$'a\db'": ((0, 'adb'),),
+            r"$'a\x1bb'": ((0, 'a\x1bb'),),
+            r"$'\u123z'": ((0, '\u0123z'),),
+            r"$'\U0001F1E8'": ((0, '\U0001F1E8'),),
+            r"$'\U1F1E8'": ((0, '\U0001F1E8'),),
+            r"$'a\U1F1E8'b": ((0, 'a\U0001F1E8b'),),
+        }.items():
+            actual = tuple(shlex_split_with_positions(q, True))
+            self.ae(expected, actual, f'Failed for text: {q!r}')
