@@ -1,11 +1,13 @@
 #!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
-from kitty.fast_data_types import DECAWM, DECCOLM, DECOM, IRM, VT_PARSER_BUFFER_SIZE, Cursor
+from kitty.config import defaults
+from kitty.fast_data_types import DECAWM, DECCOLM, DECOM, IRM, VT_PARSER_BUFFER_SIZE, Color, ColorProfile, Cursor
 from kitty.marks import marker_from_function, marker_from_regex
+from kitty.rgb import color_names
 from kitty.window import pagerhist
 
-from . import BaseTest, parse_bytes
+from . import BaseTest, draw_multicell, parse_bytes
 
 
 class TestScreen(BaseTest):
@@ -278,6 +280,74 @@ class TestScreen(BaseTest):
         self.ae(s.cursor.x, 1)
 
     def test_resize(self):
+        from kitty.window import as_text
+        def at():
+            return as_text(s, add_history=True)
+        def ac():
+            return s.line(s.cursor.y)[s.cursor.x]
+
+        # test that a wrapped line split by the history buffer is re-stitched
+        s = self.create_screen(cols=4, lines=4, scrollback=4)
+        text = ''
+        for i in range(s.lines + 1):
+            if i == 2:
+                text += 'abcd'
+            else:
+                text += str(i + 1) * s.columns
+        s.draw(text)
+        self.assertTrue(s.historybuf.endswith_wrap())
+        s.cursor.x, s.cursor.y = 1, 1
+        self.ae(ac(), 'b')
+        s.resize(s.lines, s.columns + 2)
+        self.assertTrue(s.historybuf.endswith_wrap())
+        self.ae(str(s.historybuf), '111122')
+        self.ae(at(), text + '\n')
+        self.ae((s.cursor.x, s.cursor.y), (3, 0))
+        self.ae(ac(), 'b')
+        s = self.create_screen(cols=4, lines=4, scrollback=4)
+        s.draw('1111222'), s.linefeed(), s.carriage_return()
+        s.draw('333344445555')
+        s.resize(s.lines, s.columns + 2)
+        self.ae(str(s.historybuf), '111122')
+        self.ae(str(s.line(0)), '2')
+        self.ae(at(), '1111222\n333344445555\n')
+        s = self.create_screen(cols=4, lines=4, scrollback=4)
+        s.draw('1111😸2'), s.linefeed(), s.carriage_return()
+        s.index(), s.index()
+        s.resize(s.lines, s.columns + 1)
+        self.ae(str(s.historybuf), '1111')
+        self.assertTrue(s.historybuf.endswith_wrap())
+        self.ae(at(), '1111😸2\n\n\n')
+        s = self.create_screen(cols=4, lines=4, scrollback=4)
+        s.draw(text)
+        s.cursor.x, s.cursor.y = 1, 1
+        self.ae(ac(), 'b')
+        s.resize(s.lines, s.columns * 2)
+        self.ae(ac(), 'b')
+        self.ae(str(s.historybuf), '11112222')
+        self.ae(at(), text + '\n\n')
+        self.ae((s.cursor.x, s.cursor.y), (1, 0))
+
+        # test that trailing blank line is preserved on resize
+        s = self.create_screen(cols=5, lines=5, scrollback=15)
+        for i in range(3):
+            s.draw(f'oo{i}'), s.index(), s.carriage_return()
+        s.draw('$ pp'), s.index(), s.carriage_return()
+        s.resize(s.lines, 2)
+        self.assertFalse(str(s.line(s.cursor.y)))
+        self.assertFalse(s.cursor.x)
+        # test that only happens when last line is not continued
+        s = self.create_screen(cols=5, lines=5, scrollback=15)
+        for i in range(3):
+            s.draw(f'oo{i}'), s.index(), s.carriage_return()
+        s.draw('p' * (s.columns + 2)), s.carriage_return()
+        s.resize(s.lines, 2)
+        self.assertTrue(str(s.line(s.cursor.y)))
+        s = self.create_screen(cols=5, lines=5, scrollback=15)
+        s.draw('12345'), s.carriage_return(), s.index()
+        s.resize(s.lines, s.columns - 1)
+        self.ae(('1234', '5', ''), tuple(str(s.line(i)) for i in range(s.cursor.y+1)))
+
         s = self.create_screen(scrollback=6)
         s.draw(''.join([str(i) * s.columns for i in range(s.lines)]))
         s.resize(3, 10)
@@ -290,8 +360,10 @@ class TestScreen(BaseTest):
         s = self.create_screen(scrollback=20)
         s.draw(''.join(str(i) * s.columns for i in range(s.lines*2)))
         self.ae(str(s.linebuf), '55555\n66666\n77777\n88888\n99999')
+        before = at()
         s.resize(5, 2)
-        self.ae(str(s.linebuf), '88\n88\n99\n99\n9')
+        self.ae(before, at())
+        self.ae(str(s.linebuf), '88\n88\n89\n99\n99')
         s = self.create_screen()
         s.draw('a' * s.columns)
         s.linefeed(), s.carriage_return()
@@ -304,6 +376,16 @@ class TestScreen(BaseTest):
         s.draw('x' * len(str(s.line(1))))
         s.resize(s.lines, s.columns + 4)
         self.ae(str(s.linebuf), 'xxx\nxx\nbb\n\n')
+        s = self.create_screen()
+        c = s.callbacks
+        parse_bytes(s, b'\x1b[?2048$p')  # ]
+        self.ae(c.wtcbuf, b'\x1b[?2048;2$y')  # ]
+        c.clear()
+        parse_bytes(s, b'\x1b[?2048h\x1b[?2048$p')  # ]]
+        self.ae(c.wtcbuf, b'\x1b[?2048;1$y')  # ]
+        self.ae(c.num_of_resize_events, 1)
+        parse_bytes(s, b'\x1b[?2048h')  # ]
+        self.ae(c.num_of_resize_events, 2)
 
     def test_cursor_after_resize(self):
 
@@ -555,8 +637,10 @@ class TestScreen(BaseTest):
 
         expected = ''.join(('55555', '\n66666', '\n77777', '\n88888', '\n99999'))
         self.ae(ts(), expected)
+        self.ae(ts(True), expected)
         s.scroll(2, True)
         self.ae(ts(), expected)
+        self.ae(ts(True), expected)
         s.reset()
         s.draw('ab   cd')
         s.start_selection(0, 0)
@@ -580,6 +664,7 @@ class TestScreen(BaseTest):
         self.ae(s.text_for_selection(), ('abc  ', 'xy'))
         self.ae(s.text_for_selection(True), ('a\x1b[32mb\x1b[39mc  ', 'xy', '\x1b[m'))
         self.ae(s.text_for_selection(True, True), ('a\x1b[32mb\x1b[39mc', 'xy', '\x1b[m'))
+        # ]]]]]]]]]]]]]]]]]]]]
 
     def test_soft_hyphen(self):
         s = self.create_screen()
@@ -591,17 +676,31 @@ class TestScreen(BaseTest):
 
     def test_variation_selectors(self):
         s = self.create_screen()
-        s.draw('\U0001f610')
-        self.ae(s.cursor.x, 2)
-        s.carriage_return(), s.linefeed()
-        s.draw('\U0001f610\ufe0e')
-        self.ae(s.cursor.x, 1)
-        s.carriage_return(), s.linefeed()
-        s.draw('\u25b6')
-        self.ae(s.cursor.x, 1)
-        s.carriage_return(), s.linefeed()
-        s.draw('\u25b6\ufe0f')
-        self.ae(s.cursor.x, 2)
+        def t(*a):
+            s.reset()
+            for i in range(0, len(a), 2):
+                char, x = a[i], a[i+1]
+                s.draw(char)
+                self.ae(s.cursor.x, x, f'after char: {char!r}')
+        # already wide + VS15
+        t('\U0001f610', 2, '\ufe0e', 1, '\ufe0e', 1)
+        t('\U0001f610\ufe0e', 1, '\ufe0e', 1)
+        # narrow + VS16
+        t('\u25b6', 1, '\ufe0f', 2)
+        t('\u25b6\ufe0f', 2)
+        # wide + VS16
+        t('\u26d4\ufe0f', 2, '\ufe0f', 2)
+        t('\u26d4', 2, '\ufe0f', 2)
+        # narrow + VS15
+        t('\u25b6', 1, '\ufe0e', 1)
+        t('\u25b6\ufe0e', 1)
+        # narrow + VS16 + VS15
+        t('\u25b6', 1, '\ufe0f', 2, '\ufe0e', 2)
+        # wide + VS15 + VS16
+        t('\U0001f610', 2, '\ufe0e', 1, '\ufe0f', 1)
+        # large numbers of combining chars
+        s.reset()
+        s.draw("\N{HEAVY EXCLAMATION MARK SYMBOL}" + 4500 * "\N{VARIATION SELECTOR-16}")
 
     def test_writing_with_cursor_on_trailer_of_wide_character(self):
         s = self.create_screen()
@@ -663,8 +762,8 @@ class TestScreen(BaseTest):
     def test_wrapping_serialization(self):
         from kitty.window import as_text
         s = self.create_screen(cols=2, lines=2, scrollback=2, options={'scrollback_pager_history_size': 128})
-        s.draw('aabbccddeeff')
-        self.ae(as_text(s, add_history=True), 'aabbccddeeff')
+        s.draw('ū̀abbccddeefū̀')
+        self.ae(as_text(s, add_history=True), 'ū̀abbccddeefū̀')
         self.assertNotIn('\n', as_text(s, add_history=True, as_ansi=True))
 
         s = self.create_screen(cols=2, lines=2, scrollback=2, options={'scrollback_pager_history_size': 128})
@@ -825,32 +924,32 @@ class TestScreen(BaseTest):
         s.draw('hij')
         self.ae(s.line(1).hyperlink_ids(), (1, 1, 2, 2, 2))
         set_link()
-        self.ae([('a:url-a', 1), (':url-b', 2)], s.hyperlinks_as_list())
+        self.ae({('a:url-a', 1), (':url-b', 2)}, s.hyperlinks_as_set())
         s.garbage_collect_hyperlink_pool()
-        self.ae([('a:url-a', 1), (':url-b', 2)], s.hyperlinks_as_list())
+        self.ae({('a:url-a', 1), (':url-b', 2)}, s.hyperlinks_as_set())
         for i in range(s.lines + 2):
             s.linefeed()
         s.garbage_collect_hyperlink_pool()
-        self.ae([('a:url-a', 1), (':url-b', 2)], s.hyperlinks_as_list())
+        self.ae({('a:url-a', 1), (':url-b', 2)}, s.hyperlinks_as_set())
         for i in range(s.lines * 2):
             s.linefeed()
         s.garbage_collect_hyperlink_pool()
-        self.assertFalse(s.hyperlinks_as_list())
+        self.assertFalse(s.hyperlinks_as_set())
         set_link('url-a', 'x')
         s.draw('a')
         set_link('url-a', 'y')
         s.draw('a')
         set_link()
-        self.ae([('x:url-a', 1), ('y:url-a', 2)], s.hyperlinks_as_list())
+        self.ae({('x:url-a', 1), ('y:url-a', 2)}, s.hyperlinks_as_set())
 
         s = self.create_screen()
         set_link('u' * 2048)
         s.draw('a')
-        self.ae([(':' + 'u' * 2045, 1)], s.hyperlinks_as_list())
+        self.ae({(':' + 'u' * 2045, 1)}, s.hyperlinks_as_set())
         s = self.create_screen()
         set_link('u' * 2048, 'i' * 300)
         s.draw('a')
-        self.ae([('i'*256 + ':' + 'u' * (2045 - 256), 1)], s.hyperlinks_as_list())
+        self.ae({('i'*256 + ':' + 'u' * (2045 - 256), 1)}, s.hyperlinks_as_set())
 
         s = self.create_screen()
         set_link('1'), s.draw('1')
@@ -859,19 +958,19 @@ class TestScreen(BaseTest):
         s.cursor.x = 1
         set_link(), s.draw('X')
         self.ae(s.line(0).hyperlink_ids(), (1, 0, 3, 0, 0))
-        self.ae([(':1', 1), (':2', 2), (':3', 3)], s.hyperlinks_as_list())
+        self.ae({(':1', 1), (':2', 2), (':3', 3)}, s.hyperlinks_as_set())
         s.garbage_collect_hyperlink_pool()
-        self.ae([(':1', 1), (':3', 2)], s.hyperlinks_as_list())
+        self.ae({(':1', 1), (':3', 2)}, s.hyperlinks_as_set())
         set_link('3'), s.draw('3')
-        self.ae([(':1', 1), (':3', 2)], s.hyperlinks_as_list())
+        self.ae({(':1', 1), (':3', 2)}, s.hyperlinks_as_set())
         set_link('4'), s.draw('4')
-        self.ae([(':1', 1), (':3', 2), (':4', 3)], s.hyperlinks_as_list())
+        self.ae({(':1', 1), (':3', 2), (':4', 3)}, s.hyperlinks_as_set())
 
         s = self.create_screen()
         set_link('1'), s.draw('1')
         set_link('2'), s.draw('2')
         set_link('1'), s.draw('1')
-        self.ae([(':2', 2), (':1', 1)], s.hyperlinks_as_list())
+        self.ae({(':2', 2), (':1', 1)}, s.hyperlinks_as_set())
 
         s = self.create_screen()
         set_link('1'), s.draw('12'), set_link(), s.draw('X'), set_link('1'), s.draw('3')
@@ -1002,6 +1101,7 @@ class TestScreen(BaseTest):
         def ac(idx, count):
             self.ae(c.wtcbuf, f'\033[{idx};{count}#Q'.encode('ascii'))
             c.clear()
+        # ]]]]]]]]]]]]]]]]}}}}}}}}}}}}}}}}))))))))))))))))))))))
 
         w('#R')
         ac(0, 0)
@@ -1021,40 +1121,11 @@ class TestScreen(BaseTest):
         ac(9, 10)
 
     def test_detect_url(self):
-        s = self.create_screen(cols=30)
-
-        def ae(expected, x=3, y=0):
-            s.detect_url(x, y)
-            url = ''.join(s.text_for_marked_url())
-            self.assertEqual(expected, url)
-
-        def t(url, x=0, y=0, before='', after='', expected=''):
-            s.reset()
-            s.cursor.x = x
-            s.cursor.y = y
-            s.draw(before + url + after)
-            ae(expected or url, x=x + 1 + len(before), y=y)
-
-
-        t('http://moo.com')
-        t('http://moo.com/something?else=+&what-')
-        t('http://moo.com#fragme')
-        for (st, e) in '() {} [] <>'.split():
-            t('http://moo.com', before=st, after=e)
-        for trailer in ')-=':
-            t('http://moo.com' + trailer)
-        for trailer in '{}([<>':
-            t('http://moo.com', after=trailer)
-        t('http://moo.com', x=s.columns - 9)
-        t('https://wraps-by-one-char.com', before='[', after=']')
-        t('http://[::1]:8080')
-        t('https://wr[aps-by-one-ch]ar.com')
-        t('http://[::1]:8080/x', after='[')
-        t('http://[::1]:8080/x]y34', expected='http://[::1]:8080/x')
-        t('https://wraps-by-one-char.com[]/x', after='[')
+        detect_url(self)
+        detect_url(self, scale=2)
 
     def test_prompt_marking(self):
-        s = self.create_screen()
+        # ]]]]]]]]]]]]]]]]}}}}}}}}}}}}}}}}))))))))))))))))))))))
 
         def mark_prompt():
             parse_bytes(s, b'\033]133;A\007')
@@ -1062,21 +1133,62 @@ class TestScreen(BaseTest):
         def mark_output():
             parse_bytes(s, b'\033]133;C\007')
 
+        def draw_prompt(x):
+            mark_prompt(), s.draw(f'$ {x}'), s.carriage_return(), s.index()
+
+        def draw_output(n, x='', m=True):
+            if m:
+                mark_output()
+            for i in range(n):
+                s.draw(f'{i}{x}'), s.index(), s.carriage_return()
+
+        s = self.create_screen(cols=5, lines=5, scrollback=15)
+        draw_output(3, 'oo')
+        draw_prompt('pp')
+        mark_output()
+        s.toggle_alt_screen()
+        s.resize(s.lines, 2)
+        s.toggle_alt_screen()
+        self.assertFalse(str(s.line(s.cursor.y)))
+
+        s = self.create_screen(scrollback=10)
+        draw_output(1, 'start')
         for i in range(4):
-            mark_prompt()
-            s.draw(f'$ {i}')
-            s.carriage_return()
-            s.index(), s.index()
+            draw_prompt(str(i) + 'ab')
+            draw_output(1, 'ut')
         self.ae(s.scrolled_by, 0)
+        self.ae(str(s.visual_line(0)), '$ 2ab')
+        self.assertFalse(s.scroll_to_prompt(-3))
+        self.assertTrue(s.scroll_to_prompt())
+        self.ae(str(s.visual_line(0)), '$ 1ab')
+        self.assertTrue(s.scroll_to_prompt())
+        self.ae(str(s.visual_line(0)), '$ 0ab')
+        self.assertFalse(s.scroll_to_prompt())
+        self.assertFalse(s.scroll_to_prompt(4))
+        self.assertTrue(s.scroll_to_prompt(1))
+        self.ae(str(s.visual_line(0)), '$ 1ab')
+        self.assertTrue(s.scroll_to_prompt(1))
+        self.ae(str(s.visual_line(0)), '$ 2ab')
+        self.assertFalse(s.scroll_to_prompt(1))
+        # wrap prompts
+        s.resize(s.lines + 1, s.columns - 2)
+        self.ae(s.scrolled_by, 0)
+        self.ae(str(s.visual_line(0)), 'ab')
+        self.assertFalse(s.scroll_to_prompt(-4))
+        self.assertTrue(s.scroll_to_prompt())
+        self.ae(str(s.visual_line(0)), '$ 2')
         self.assertTrue(s.scroll_to_prompt())
         self.ae(str(s.visual_line(0)), '$ 1')
         self.assertTrue(s.scroll_to_prompt())
         self.ae(str(s.visual_line(0)), '$ 0')
         self.assertFalse(s.scroll_to_prompt())
+        self.assertFalse(s.scroll_to_prompt(4))
         self.assertTrue(s.scroll_to_prompt(1))
         self.ae(str(s.visual_line(0)), '$ 1')
         self.assertTrue(s.scroll_to_prompt(1))
         self.ae(str(s.visual_line(0)), '$ 2')
+        self.assertTrue(s.scroll_to_prompt(1))
+        self.ae(str(s.visual_line(0)), 'ab')
         self.assertFalse(s.scroll_to_prompt(1))
 
         s = self.create_screen()
@@ -1107,6 +1219,14 @@ class TestScreen(BaseTest):
             s.cmd_output(2, a.append)
             return ''.join(a)
 
+        s = self.create_screen(cols=5, lines=5, scrollback=15)
+        draw_prompt('P' * s.columns)
+        draw_output(s.lines + 1, 'a')  # ensure prompt is in scrollback
+        draw_prompt('Q' * s.columns)
+        draw_output(s.lines + 1, 'b')  # ensure prompt is in scrollback
+        draw_prompt('R' * s.columns)
+        self.ae(lco(), '0b\n1b\n2b\n3b\n4b\n5b')
+
         s = self.create_screen()
         s.draw('abcd'), s.index(), s.carriage_return()
         s.draw('12'), s.index(), s.carriage_return()
@@ -1121,16 +1241,7 @@ class TestScreen(BaseTest):
         mark_prompt(), s.draw('$ 1')
         self.ae(fco(), 'abcd\n12')
         self.ae(lco(), 'abcd\n12')
-        self.ae(lco(as_ansi=True), '\x1b[m\x1b]133;C\x1b\\abcd\n\x1b[m12')
-
-        def draw_prompt(x):
-            mark_prompt(), s.draw(f'$ {x}'), s.carriage_return(), s.index()
-
-        def draw_output(n, x='', m=True):
-            if m:
-                mark_output()
-            for i in range(n):
-                s.draw(f'{i}{x}'), s.index(), s.carriage_return()
+        self.ae(lco(as_ansi=True), '\x1b[m\x1b]133;C\x1b\\abcd\n\x1b[m12')  # ]]]
 
         s = self.create_screen(cols=5, lines=5, scrollback=15)
         draw_output(1, 'start', False)
@@ -1159,10 +1270,19 @@ class TestScreen(BaseTest):
         s.scroll(2, False)
         self.ae(str(s.visual_line(0)), '$ 1')
         self.ae(fco(), '0x\n1x\n')
+        # test: obscure prompt in scrollback
+        s.resize(3, 5)
+        self.ae(str(s.visual_line(0)), '0x')
+        # without succeeding prompt
+        self.ae(fco(), '0x\n1x\n')
+        s.resize(4, 5)
+        draw_prompt('3')
+        # with succeeding prompt
+        self.ae(fco(), '0x\n1x')
 
         # resize
         # get last cmd output with continued output mark
-        draw_prompt('3'), draw_output(1, 'long_line'), draw_output(2, 'l', False)
+        draw_output(1, 'long_line'), draw_output(2, 'l', False)
         s.resize(4, 5)
         s.scroll_to_prompt(-4)
         self.ae(str(s.visual_line(0)), '$ 0')
@@ -1172,6 +1292,58 @@ class TestScreen(BaseTest):
         self.ae(lvco(), '0\n1\n2')
         s.scroll_to_prompt(1)
         self.ae(lvco(), '0x\n1x')
+        # test: obscure prompt
+        s.scroll(2, False)
+        self.ae(lvco(), '0x\n1x')
+        # test: prompts without output
+        s.scroll(s.scrolled_by, False)
+        draw_prompt('4')
+        # resizing resets last visited, rely on scroll_to_prompt
+        s.resize(2, s.columns + 4)
+        s.scroll_to_prompt()
+        s.scroll_to_prompt(1)
+        self.ae(str(s.visual_line(0)), '$ 4')
+        self.ae(lvco(), '')
+        s.resize(5, s.columns)
+        draw_prompt('wrapcmd')
+        s.scroll_to_prompt(1)
+        self.ae(lvco(), '')
+        draw_output(1, 'wrapout'), draw_output(1, 'y', False)
+        self.ae(lvco(), '0wrapout\n0y\n')
+        # wrap long prompt with long output
+        s.resize(5, s.columns - 4)
+        self.ae(str(s.visual_line(0)), 'pcmd')
+        # test: set last visited to previous empty prompt
+        s.scroll_to_prompt(-2)
+        self.ae(str(s.visual_line(0)), '$ 4')
+        self.ae(lvco(), '0wrapout\n0y\n')
+        mark_prompt(), s.draw('$ end')
+        self.ae(lvco(), '0wrapout\n0y')
+        # test: set last visited to continued line of long prompt
+        s.scroll_to_prompt(2)
+        self.ae(str(s.visual_line(0)), 'pcmd')
+        self.ae(lvco(), '0wrapout\n0y')
+        s.scroll_to_prompt()
+        self.ae(lvco(), '0wrapout\n0y')
+        # test: set last visited to continued line of output
+        s.carriage_return(), s.index()
+        draw_output(1, 'z')
+        s.scroll_to_prompt(1)
+        self.ae(lvco(), '0wrapout\n0y')
+
+        # test that post rewrap prompt lines have correct attributes
+        s = self.create_screen(cols=5, lines=5, scrollback=15)
+        draw_prompt('P' * (s.columns - 2))
+        draw_output(s.lines + 1, 'a')  # ensure prompt is in scrollback
+        draw_prompt('Q' * (s.columns - 2))
+        self.ae(str(s.visual_line(0)), '3a')
+        s.scroll_to_prompt()
+        self.ae(str(s.visual_line(0)), '$ PPP')
+        s.scroll_to_prompt(1)
+        self.ae(str(s.visual_line(0)), '3a')
+        s.resize(s.lines, s.columns - 2)
+        s.scroll_to_prompt()
+        self.ae(str(s.visual_line(0)), '$ P')
 
         # last command output without line break
         s = self.create_screen(cols=10, lines=3)
@@ -1248,3 +1420,69 @@ class TestScreen(BaseTest):
         t('<', '0')
         t('=left_ptr', 'default')
         t('=fleur', 'move')
+
+    def test_color_profile(self):
+        c = ColorProfile(defaults)
+        for i in range(8):
+            col = getattr(defaults, f'color{i}')
+            self.ae(c.as_color(i << 8 | 1), col)
+        self.ae(c.as_color(255 << 8 | 1), Color(0xee, 0xee, 0xee))
+        s = self.create_screen()
+        s.color_profile.reload_from_opts(defaults)
+        def q(send, expected=None):
+            s.callbacks.clear()
+            parse_bytes(s, b'\x1b]21;' + ';'.join(f'{k}={v}' for k, v in send.items()).encode() + b'\a')
+            self.ae(s.callbacks.color_control_responses, [expected] if expected else [])
+        q({k: '?' for k in 'background foreground 213 unknown'.split()}, {
+            'background': defaults.background, 'foreground': defaults.foreground, '213': defaults.color213, 'unknown': '?'})
+        q({'background':'aquamarine'})
+        q({'background':'?', 'selection_background': '?'}, {'background': color_names['aquamarine'], 'selection_background': s.color_profile.highlight_bg})
+        q({'selection_background': ''})
+        self.assertIsNone(s.color_profile.highlight_bg)
+        q({'selection_background': '?'}, {'selection_background': ''})
+        s.color_profile.reload_from_opts(defaults)
+        q({'transparent_background_color9': '?'}, {'transparent_background_color9': '?'})
+        q({'transparent_background_color2': '?'}, {'transparent_background_color2': ''})
+        q({'transparent_background_color2': 'red@0.5'})
+        q({'transparent_background_color2': '?'}, {'transparent_background_color2': (Color(255, 0, 0), 126)})
+        q({'transparent_background_color2': '#ffffff@-1'})
+        q({'transparent_background_color2': '?'}, {'transparent_background_color2': (Color(255, 255, 255), 255)})
+
+
+def detect_url(self, scale=1):
+    s = self.create_screen(cols=30 * scale)
+
+    def ae(expected, x=3, y=0):
+        s.detect_url(x * scale, y * scale)
+        url = ''.join(s.text_for_marked_url())
+        self.assertEqual(expected, url)
+
+    def t(url, x=0, y=0, before='', after='', expected=''):
+        s.reset()
+        s.cursor.x = x
+        s.cursor.y = y
+        text = before + url + after
+        if scale == 1:
+            s.draw(text)
+        else:
+            draw_multicell(s, text, scale=scale)
+        ae(expected or url, x=x + 1 + len(before), y=y)
+
+
+    t('http://moo.com')
+    t('http://moo.com/something?else=+&what-')
+    t('http://moo.com#fragme')
+    for (st, e) in '() {} [] <>'.split():
+        t('http://moo.com', before=st, after=e)
+    for trailer in ')-=':
+        t('http://moo.com' + trailer)
+    for trailer in '{}([<>':   # )]>
+        t('http://moo.com', after=trailer)
+    if scale == 1:
+        t('http://moo.com', x=s.columns - 9)
+    t('https://wraps-by-one-char.com', before='[', after=']')
+    t('http://[::1]:8080')
+    t('https://wr[aps-by-one-ch]ar.com')
+    t('http://[::1]:8080/x', after='[')  # ]
+    t('http://[::1]:8080/x]y34', expected='http://[::1]:8080/x')
+    t('https://wraps-by-one-char.com[]/x', after='[')  # ]
