@@ -12,14 +12,14 @@ from contextlib import contextmanager
 from functools import lru_cache, partial
 
 from kitty.bash import decode_ansi_c_quoted_string
-from kitty.constants import kitten_exe, kitty_base_dir, shell_integration_dir, terminfo_dir
+from kitty.constants import is_macos, kitten_exe, kitty_base_dir, shell_integration_dir, terminfo_dir
 from kitty.fast_data_types import CURSOR_BEAM, CURSOR_BLOCK, CURSOR_UNDERLINE
 from kitty.shell_integration import setup_bash_env, setup_fish_env, setup_zsh_env
 
 from . import BaseTest
 
 
-@lru_cache()
+@lru_cache
 def bash_ok():
     v = shutil.which('bash')
     if not v:
@@ -94,7 +94,7 @@ class ShellIntegration(BaseTest):
         try:
             if self.with_kitten:
                 cmd = [kitten_exe(), 'run-shell', '--shell', shlex.join(cmd)]
-            pty = self.create_pty(cmd, cwd=home_dir, env=env)
+            pty = self.create_pty(cmd, cwd=home_dir, env=env, cols=180)
             i = 10
             while i > 0 and not pty.screen_contents().strip():
                 pty.process_input_from_child()
@@ -121,6 +121,7 @@ RPS1="{rps1}"
             self.ae(pty.callbacks.titlebuf[-1], '~')
             pty.callbacks.clear()
             pty.send_cmd_to_child('mkdir test && ls -a')
+            self.assert_command(pty)
             pty.wait_till(lambda: pty.screen_contents().count(rps1) == 2)
             self.ae(pty.callbacks.titlebuf[-2:], ['mkdir test && ls -a', '~'])
             q = '\n'.join(str(pty.screen.line(i)) for i in range(1, pty.screen.cursor.y))
@@ -138,10 +139,12 @@ RPS1="{rps1}"
             pty.wait_till(redrawn)
             self.ae(q, str(pty.screen.line(pty.screen.cursor.y)))
             pty.write_to_child('\r')
+            self.assert_command(pty, 'echo $COLUMNS')
             pty.wait_till(lambda: pty.screen_contents().count(rps1) == 3)
             self.ae('40', str(pty.screen.line(pty.screen.cursor.y - 1)))
             self.ae(q, str(pty.screen.line(pty.screen.cursor.y - 2)))
             pty.send_cmd_to_child('clear')
+            self.assert_command(pty)
             q = ps1 + ' ' * (pty.screen.columns - len(ps1) - len(rps1)) + rps1
             pty.wait_till(lambda: pty.screen_contents() == q)
             pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BEAM)
@@ -149,6 +152,7 @@ RPS1="{rps1}"
             pty.wait_till(lambda: pty.screen.cursor.shape == 0)
             pty.write_to_child('\x04')
             pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BEAM)
+            self.assert_command(pty)
         with self.run_shell(rc=f'''PS1="{ps1}"''') as pty:
             pty.callbacks.clear()
             pty.send_cmd_to_child('printf "%s\x16\a%s" "a" "b"')
@@ -159,10 +163,13 @@ RPS1="{rps1}"
             os.mkdir(q)
             pty.send_cmd_to_child(f'cd {q}')
             pty.wait_till(lambda: pty.screen.last_reported_cwd.decode().endswith(q))
+            if not is_macos:  # Fails on older macOS like the one used to build kitty binary because of unicode encoding issues
+                self.assert_command(pty)
         with self.run_shell(rc=f'''PS1="{ps1}"\nexport ES="a\n b c\nd"''') as pty:
             pty.callbacks.clear()
             pty.send_cmd_to_child('clone-in-kitty')
             pty.wait_till(lambda: len(pty.callbacks.clone_cmds) == 1)
+            self.assert_command(pty)
             env = pty.callbacks.clone_cmds[0].env
             self.ae(env.get('ES'), 'a\n b c\nd')
 
@@ -184,15 +191,18 @@ function _set_status_prompt; function fish_prompt; echo -n "$pipestatus $status
             pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 1)
             self.ae(pty.screen_contents(), q)
 
-            # shell integration dir must no be in XDG_DATA_DIRS
-            pty.send_cmd_to_child(f'string match -q -- "*{shell_integration_dir}*" "$XDG_DATA_DIRS" || echo "XDD_OK"')
+            # shell integration dir must not be in XDG_DATA_DIRS
+            cmd = f'string match -q -- "*{shell_integration_dir}*" "$XDG_DATA_DIRS" || echo "XDD_OK"'
+            pty.send_cmd_to_child(cmd)
             pty.wait_till(lambda: 'XDD_OK' in pty.screen_contents())
+            # self.assert_command(pty, cmd)
 
             # CWD reporting
             self.assertTrue(pty.screen.last_reported_cwd.decode().endswith(self.home_dir))
             q = os.path.join(self.home_dir, 'testing-cwd-notification-🐱')
             os.mkdir(q)
             pty.send_cmd_to_child(f'cd {q}')
+            # self.assert_command(pty)
             pty.wait_till(lambda: pty.screen.last_reported_cwd.decode().endswith(q))
             pty.send_cmd_to_child('cd -')
             pty.wait_till(lambda: pty.screen.last_reported_cwd.decode().endswith(self.home_dir))
@@ -202,6 +212,7 @@ function _set_status_prompt; function fish_prompt; echo -n "$pipestatus $status
             pty.send_cmd_to_child('clear')
             pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 1)
             pty.send_cmd_to_child('_test_comp_path')
+            # self.assert_command(pty)
             pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 2)
             q = '\n'.join(str(pty.screen.line(i)) for i in range(1, pty.screen.cursor.y))
             self.ae(q, 'ok')
@@ -221,6 +232,7 @@ function _set_status_prompt; function fish_prompt; echo -n "$pipestatus $status
             self.ae(q, str(pty.screen.line(pty.screen.cursor.y)))
             pty.write_to_child('\r')
             pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 3)
+            # self.assert_command(pty, 'echo $COLUMNS')
             self.ae('40', str(pty.screen.line(pty.screen.cursor.y - 1)))
             self.ae(q, str(pty.screen.line(pty.screen.cursor.y - 2)))
 
@@ -245,11 +257,17 @@ function _set_status_prompt; function fish_prompt; echo -n "$pipestatus $status
             pty.write_to_child('i')
             pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BEAM)
             pty.send_cmd_to_child('_set_key default')
+            # self.assert_command(pty)
             pty.wait_till(lambda: pty.screen_contents().count(right_prompt) == 4)
             pty.wait_till(lambda: pty.screen.cursor.shape == CURSOR_BEAM)
 
             pty.send_cmd_to_child('exit')
 
+    def assert_command(self, pty, cmd='', exit_status=0):
+        cmd = cmd or pty.last_cmd
+        pty.wait_till(lambda: pty.callbacks.last_cmd_exit_status == 0, timeout_msg=lambda: f'{pty.callbacks.last_cmd_exit_status=} != {exit_status}')
+        pty.wait_till(lambda: pty.callbacks.last_cmd_cmdline == cmd, timeout_msg=lambda: f'{pty.callbacks.last_cmd_cmdline=!r} != {cmd!r}')
+
     @unittest.skipUnless(bash_ok(), 'bash not installed, too old, or debug build')
     def test_bash_integration(self):
         ps1 = 'prompt> '
@@ -266,9 +284,11 @@ PS1="{ps1}"
             pty.wait_till(lambda: pty.callbacks.titlebuf[-1:] == ['~'])
             self.ae(pty.callbacks.titlebuf[-1], '~')
             pty.callbacks.clear()
-            pty.send_cmd_to_child('mkdir test && ls -a')
-            pty.wait_till(lambda: pty.callbacks.titlebuf[-2:] == ['mkdir test && ls -a', '~'])
+            cmd = 'mkdir test && ls -a'
+            pty.send_cmd_to_child(cmd)
+            pty.wait_till(lambda: pty.callbacks.titlebuf[-2:] == [cmd, '~'])
             pty.wait_till(lambda: pty.screen_contents().count(ps1) == 2)
+            self.assert_command(pty, cmd)
             q = '\n'.join(str(pty.screen.line(i)) for i in range(1, pty.screen.cursor.y))
             self.ae(pty.last_cmd_output(), q)
             # shrink the screen
@@ -300,6 +320,7 @@ PS1="{ps1}"
             pty.callbacks.clear()
             pty.send_cmd_to_child('declare')
             pty.wait_till(lambda: 'LOCAL_KSI_VAR' in pty.screen_contents())
+            self.assert_command(pty, 'declare')
         with self.run_shell(shell='bash', rc=f'''PS1="{ps1}"''') as pty:
             pty.callbacks.clear()
             pty.send_cmd_to_child('printf "%s\x16\a%s" "a" "b"')
@@ -332,6 +353,7 @@ PS1="{ps1}"
                 pty.wait_till(lambda: pty.screen_contents().count(ps1) == 3)
                 self.ae('40', str(pty.screen.line(pty.screen.cursor.y - len(ps1.splitlines()))))
                 self.ae(ps1.splitlines()[-1] + 'echo $COLUMNS', str(pty.screen.line(pty.screen.cursor.y - 1 - len(ps1.splitlines()))))
+                self.assert_command(pty, 'echo $COLUMNS')
 
         # test startup file sourcing
 
@@ -382,6 +404,7 @@ PS1="{ps1}"
             env = pty.callbacks.clone_cmds[0].env
             self.ae(env.get('ES'), 'a\n `b` c\n$d', f'Screen contents: {pty.screen_contents()!r}')
             self.ae(env.get('ES2'), 'XXX', f'Screen contents: {pty.screen_contents()!r}')
+
         for q, e in {
             'a': 'a',
             r'a\ab': 'a\ab',
@@ -390,8 +413,7 @@ PS1="{ps1}"
             r'a\U1f345x': 'a🍅x',
             r'a\c b': 'a\0b',
         }.items():
-            q = q + "'"
-            self.ae(decode_ansi_c_quoted_string(q, 0)[0], e, f'Failed to decode: {q!r}')
+            self.ae(decode_ansi_c_quoted_string(f"$'{q}'"), e, f'Failed to decode: {q!r}')
 
 
 class ShellIntegrationWithKitten(ShellIntegration):
